#include "Pkt.h"

static int      Zero = 0;

#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <synch.h>
#include <ctype.h>

#include "swapbytes.h"
#include "tr.h"
#include "coords.h"
#include "stock.h"

#define MAX_CH_SIZE 1000000
#define MAX_STASH_SIZE 1000000
#define MAX_NSAMP  10000
#define MAX_SEED_NSAMP  6700	       /* for a 4096 byte record */
#define MAX_CD1_NSAMP 20000
#define GCF_MOTOROLA_BYTEORDER 1
#define GCF_INTEL_BYTEORDER 2
#define GCF_MAX_NREC 250
#define MAX_24BIT_INT 0x007FFFFF
#define MIN_24BIT_INT ( - MAX_24BIT_INT )
#define SECONDS_PER_DAY (24*60*60)

typedef unsigned char uchar;

static Arr     *schema_arr = 0;
static mutex_t  Database_mutex = DEFAULTMUTEX;

extern void     HD2NUC (char *, double *, int);
extern void     NUC2HD (double *, char *, int);
extern void     NUS2HI (int *, char *, int);
extern void     HI2NUS (char *, int *, int);
extern void     IUS2HI (int *, char *, int);
extern void     HI2IUS (char *, int *, int);
extern void     NUC2HI (int *, char *, int);
extern void     HI2NUC (char *, int *, int);
extern void     IS2HI (int *, char *, int);
extern void     HI2IS (char *, int *, int);
extern void     II2HI (int *, char *, int);
extern void     HI2II (char *, int *, int);
extern void     IUI2HUI (unsigned int *, char *, int);
extern void     HUI2IUI (char *, unsigned int *, int);

void
HUI2IUI (char *to, unsigned int *from, int n)
{
    int             index;
    unsigned int    i;
    for (index = 0; index < n; index++) {
	i = from[index];
	H2N4 (to, (char *) &i, 1);
	swap4 (to, to, 1);
	to += sizeof (unsigned int);
    }
}

void
IUI2HUI (unsigned int *to, char *from, int n)
{
    int             index;
    unsigned int    i;
    for (index = 0; index < n; index++) {
	swap4 (from, from, 1);
	N2H4 ((char *) &i, from, 1);
	to[index] = i;
	from += sizeof (unsigned int);
    }
}

void
HI2II (char *to, int *from, int n)
{
    int             index;
    int             i;
    for (index = 0; index < n; index++) {
	i = from[index];
	H2N4 (to, (char *) &i, 1);
	swap4 (to, to, 1);
	to += sizeof (int);
    }
}

void
II2HI (int *to, char *from, int n)
{
    int             index;
    int             i;
    for (index = 0; index < n; index++) {
	swap4 (from, from, 1);
	N2H4 ((char *) &i, from, 1);
	to[index] = i;
	from += sizeof (int);
    }
}

void
HI2IS (char *to, int *from, int n)
{
    int             i;
    short           s;
    for (i = 0; i < n; i++) {
	s = from[i];
	H2N2 (to, (char *) &s, 1);
	swap2 (to, to, 1);
	to += sizeof (short);
    }
}

void
IS2HI (int *to, char *from, int n)
{
    int             i;
    short           s;
    for (i = 0; i < n; i++) {
	swap2 (from, from, 1);
	N2H2 ((char *) &s, from, 1);
	to[i] = s;
	from += sizeof (short);
    }
}

void
HD2NUC (char *to, double *from, int n)
{
    int             i;
    unsigned char   c;
    for (i = 0; i < n; i++) {
	c = from[i];
	to[i] = c;
    }
}

void
NUC2HD (double *to, char *from, int n)
{
    int             i;
    unsigned char   c;
    for (i = 0; i < n; i++) {
	c = from[i];
	to[i] = c;
    }
}

void
IUS2HI (int *to, char *from, int n)
{
    int             i;
    unsigned short  s;
    for (i = 0; i < n; i++) {
	swap2 (from, from, 1);
	N2H2 ((char *) &s, from, 1);
	to[i] = s;
	from += sizeof (unsigned short);
    }
}

void
NUS2HI (int *to, char *from, int n)
{
    int             i;
    unsigned short  s;
    for (i = 0; i < n; i++) {
	N2H2 ((char *) &s, from, 1);
	to[i] = s;
	from += sizeof (unsigned short);
    }
}

void
HI2NUS (char *to, int *from, int n)
{
    int             i;
    unsigned short  s;
    for (i = 0; i < n; i++) {
	s = from[i];
	H2N2 (to, (char *) &s, 1);
	to += sizeof (unsigned short);
    }
}

void
HI2IUS (char *to, int *from, int n)
{
    int             i;
    unsigned short  s;
    for (i = 0; i < n; i++) {
	s = from[i];
	H2N2 (to, (char *) &s, 1);
	swap2 (to, to, 1);
	to += sizeof (unsigned short);
    }
}

void
NUC2HI (int *to, char *from, int n)
{
    int             i;
    unsigned char   c;
    for (i = 0; i < n; i++) {
	c = (unsigned char) from[i];
	to[i] = c;
    }
}

void
HI2NUC (char *to, int *from, int n)
{
    int             i;
    unsigned char   c;
    for (i = 0; i < n; i++) {
	c = from[i];
	to[i] = (char) c;
    }
}

static void
zstrcopy (char *dest, char *source, int n)
{
    int             i = 0;
    for (i = 0; i < n; i++) {
	if (*source == 0)
	    break;
	*dest++ = *source++;
    }
    for (; i < n; i++) {
	*dest++ = 0;
    }
}

static int
bad_ascii (char *s, int n)
{
    int             i,
                    c;
    for (i = 0; i < n; i++) {
	c = s[i];
	if (c == 0)
	    break;
	if (c == '\\') {
	    i++;
	} else if (!(isprint (c) || isspace (c))) {
	    return -1;
	}
    }
    if (i == n)
	return -1;
    return 0;
}


static          Dbptr
temp_database (char *schema)
{
    Dbptr           db,
                   *dbp;
    if (schema_arr == 0) {
	schema_arr = newarr (0);
    }
    dbp = (Dbptr *) getarr (schema_arr, schema);
    if (dbp == 0) {
	db = dbtmp (schema);
	setarr (schema_arr, schema, memdup (&db, sizeof (Dbptr)));
    } else {
	db = *dbp;
    }
    return db;
}

static int
unsteim (int *data, int nsamp, char *cp, int size, int level)
{
    Msd            *msd;
    int            *msd_data;
    int             retcode,
                    npts;
    if (level == 0) {
	vi2hi (&cp, data, nsamp);
	retcode = 0;
    } else {
	msd = msdnew ();
	msdput (msd,
		MSD_RECORD, cp,
		MSD_RECORD_SIZE, size,
		MSD_FIRST_DATA_FRAME, 0,
		MSD_NSAMP, nsamp,
		MSD_LEVEL, level,
		0);
	retcode = ustc (msd, &msd_data, &npts);
	memcpy (data, msd_data, npts * sizeof (int));
	msdfree (msd);
    }
    return retcode;
}

static int
deseed (char *cp, int nbytes, double *time, double *samprate,
	int *nsamp, int **data, int *datasz, Hook ** hook)
{
    Msd            *msd;
    int            *msd_data,
                    record_size,
                    npts,
                    retcode;
    if (*hook == 0 || (*hook)->free != msdfree) {
	if (*hook != 0) {
	    free_hook (hook);
	}
	*hook = new_hook ((void (*) (void *)) msdfree);
	msd = msdnew ();
	(*hook)->p = msd;
    } else {
	msd = (Msd *) (*hook)->p;
    }
    msdput (msd,
	    MSD_RECORD, cp,
	    MSD_RECORD_SIZE, nbytes,
	    0);
    retcode = msdhdr_unpack (msd);
    if (retcode == 0) {
	msdget (msd,
		MSD_RECORD_SIZE, &record_size,
		MSD_NSAMP, nsamp,
		MSD_SAMPRATE, samprate,
		MSD_TIME, time,
		0);
	if (nbytes < record_size) {
	    retcode = -1;
	} else if (*nsamp == 0 || *samprate == 0.0) {
	    retcode = Pkt_ch;
	} else {
	    retcode = umsd (msd, &msd_data, &npts);
	    SIZE_BUFFER (int *, *data, *datasz, npts);
	    memcpy (*data, msd_data, npts * sizeof (int));
	}
    }
    return retcode;
}

static int
uncanada (int *data, int nsamp, int datasz, char *cp, int nbytes)
{
    int             retcode,
                    ns;
    int            *dat = NULL;
    int             datsz = 0;
    int             keep = 0;

    if (nsamp % 20) {
	ns = ((nsamp - 1) / 20 + 1) * 20;
    } else {
	ns = nsamp;
	dat = data;
	datsz = datasz;
	keep = 1;
    }
    retcode = ucanada (cp, nbytes, &dat, &ns, &datsz);
    if (retcode < 0) {
	return retcode;
    }
    if (keep && (dat != data || datsz != datasz)) {
	register_error (0, "uncanada: Attempt to realloc fixed buffer.");
	return (-1);
    }
    if (dat != data) {
	memcpy (data, dat, nsamp * sizeof (int));
	free (dat);
    }
    return retcode;
}

static int
cmcanada (int *data, int nsamp, char *cp, int *nbytes)
{
    int             retcode;
    int            *dat;
    int             ns,
                    nb,
                    sz;
    char           *cpt;

    if (nsamp % 20) {
	ns = ((nsamp - 1) / 20 + 1) * 20;
	dat = (int *) malloc (ns * sizeof (int));
	if (dat == NULL) {
	    register_error (1, "cmcanada: malloc() error.");
	    return (-1);
	}
	memset (dat, 0, ns * sizeof (int));
	memcpy (dat, data, nsamp * sizeof (int));
    } else {
	ns = nsamp;
	dat = data;
    }
    cpt = NULL;
    nb = 0;
    sz = 0;
    retcode = ccanada (dat, ns, &cpt, &nb, &sz);
    if (retcode < 0) {
	register_error (0, "cmcanada: ccanada() error.");
	return retcode;
    }
    if (nb > *nbytes) {
	register_error (1, "cmcanada: packet buffer too small (needs %d, has %d).", nb, *nbytes);
	return (-1);
    }
    *nbytes = nb;
    memcpy (cp, cpt, nb);
    free (cpt);
    if (dat != data)
	free (dat);

    return retcode;
}

static double
dbtime (Dbptr db)
{
    double          time;
    int             mark;
    mark = elog_mark ();
    if (dbgetv (db, 0, "time", &time, 0) != 0) {
	if (dbgetv (db, 0, "lddate", &time, 0) != 0) {
	    time = now ();
	}
    }
    elog_flush (0, mark);
    return time;
}

static int
upscl (char *record, int nsamp, int *data)
{
    Msd            *msd;
    int            *sud;
    int             npts,
                    nerr;
    npts = 0;
    msd = msdnew ();
    msdput (msd,
	    MSD_LEVEL, 1,
	    MSD_FIRST_DATA_FRAME, 1,
	    MSD_NSAMP, nsamp,
	    MSD_RECORD_SIZE, 1024,
	    MSD_RECORD, record,
	    0);
    if (nerr = ustc (msd, &sud, &npts)) {
	msdfree (msd);
	complain (0, "Steim decompression reports %d errors", nerr);
	return -1;
    }
    memcpy (data, sud, nsamp * sizeof (int));
    msdfree (msd);
    return 0;
}

static double
datecode2epoch (int datecode)
{
    double          epoch;
    int             days_since_Nov_17_1989;
    int             seconds_since_midnight;

    epoch = 627264000.0;	       /* Seconds from Midnight 1/1/70 to
				        * Nov. 17, 1989 */

    days_since_Nov_17_1989 = datecode >> 17;
    seconds_since_midnight = datecode & 0x0001FFFF;

    epoch += days_since_Nov_17_1989 * SECONDS_PER_DAY;
    epoch += seconds_since_midnight;

    return epoch;
}

static char    *
decodebase36 (unsigned int v, char *s, int space)
{
    int             imed;
    int             i;
    char           *c;
    int             len;

    memset (s, 0, space + 1);

    for (i = space - 1; i >= 0; i--) {
	imed = v % 36;
	if (imed <= 9) {
	    s[i] = imed + '0';
	} else {
	    s[i] = imed - 10 + 'A';
	}
	v /= 36;
    }

    c = &s[0];
    while (*c == '0')
	c++;
    len = strlen (s);
    strncpy (s, c, len);

    return (s);
}

static int
add24bit (int val, int diff)
{
    int             res;

    res = val + diff;

    /* If the result goes outside 24 bit range, change sign of second value,
     * and repeat the add operation. This is required since two 24-bit
     * numbers can have a 25-bit difference. Only 24 bits are transmitted, so
     * the sign it is lost. This must be determined at decompression time */

    if ((res > MAX_24BIT_INT) || (res < MIN_24BIT_INT)) {
	diff = diff ^ 0xFF000000;
	res = val + diff;
    }
    return res;
}



int
stuff_junk (Packet * pkt, char *srcname, double *opkttime, char **ppp, int *nbytes, int *ppsz)
{
    char           *pp,
                   *cp;
    int             retcode = 0;
    pp = cp = *ppp;
    RESIZE_BUFFER (char *, *ppp, *ppsz, 10);
    cp = *ppp + (cp - pp);
    pp = *ppp;
    retcode = -1;
    *nbytes = cp - pp;
    insist (*nbytes < 10);
    return retcode;
}



int
unstuff_junk (char *srcname, double ipkttime, char *packet, int nbytes, Packet * pkt)
{
    char           *pp,
                   *cp;
    int             retcode = 0;
    pp = cp = packet;
    retcode = -1;
    if (nbytes < cp - pp) {
	complain (0, "nbytes=%d < cp-pp=%d\n", nbytes, cp - pp);
	retcode = -1;
    };
    return retcode;
}


int
stuff_dbug (Packet * pkt, char *srcname, double *opkttime, char **ppp, int *nbytes, int *ppsz)
{
    char           *pp,
                   *cp;
    int             retcode = 0;
    int             dummy = 0;
    int             ICHAN = 0;
    PktChannel     *achan;
    achan = gettbl (pkt->channels, 0);
    pp = cp = *ppp;
    RESIZE_BUFFER (char *, *ppp, *ppsz, (20 + (36 + 3 * PKT_NAMESIZE + achan->nsamp * 4) * pkt->nchannels));
    cp = *ppp + (cp - pp);
    pp = *ppp;
    HI2NC (cp, &pkt->version, 1);
    cp += 1 * 1;
    switch (pkt->version) {
      case 0:
	memset (cp, 0, 7);
	cp += 7;
	break;
    }
    *opkttime = achan->time;
    H2N4 (cp, (char *) &pkt->nchannels, 1);
    cp += 4 * 1;
    for (ICHAN = 0; ICHAN < pkt->nchannels; ICHAN++) {
	achan = gettbl (pkt->channels, ICHAN);
	if (achan == 0) {
	    achan = newPktChannel ();
	    pushtbl (pkt->channels, achan);
	};
	memcpy (cp, &achan->net, PKT_NAMESIZE);
	cp += 1 * PKT_NAMESIZE;
	memcpy (cp, &achan->sta, PKT_NAMESIZE);
	cp += 1 * PKT_NAMESIZE;
	memcpy (cp, &achan->chan, PKT_NAMESIZE);
	cp += 1 * PKT_NAMESIZE;
	H2ND ((uchar *) cp, &achan->time, 1);
	cp += 8 * 1;
	H2ND ((uchar *) cp, &achan->samprate, 1);
	cp += 8 * 1;
	H2ND ((uchar *) cp, &achan->calib, 1);
	cp += 8 * 1;
	switch (pkt->version) {
	  case 0:
	    H2N4 (cp, (char *) &achan->nsamp, 1);
	    cp += 4 * 1;
	    H2N4 (cp, (char *) &dummy, 1);
	    cp += 4 * 1;
	    H2N4 (cp, (char *) &dummy, 1);
	    cp += 4 * 1;
	    break;
	  default:
	    H2ND ((uchar *) cp, &achan->calper, 1);
	    cp += 8 * 1;
	    memcpy (cp, &achan->segtype, 2);
	    cp += 1 * 2;
	    H2N4 (cp, (char *) &achan->nsamp, 1);
	    cp += 4 * 1;
	    break;
	}
    };
    for (ICHAN = 0; ICHAN < pkt->nchannels; ICHAN++) {
	achan = gettbl (pkt->channels, ICHAN);
	if (achan == 0) {
	    achan = newPktChannel ();
	    pushtbl (pkt->channels, achan);
	};
	H2N4 (cp, (char *) achan->data, achan->nsamp);
	cp += 4 * achan->nsamp;
    };
    *nbytes = cp - pp;
    insist (*nbytes < (20 + (36 + 3 * PKT_NAMESIZE + achan->nsamp * 4) * pkt->nchannels));
    return retcode;
}



int
unstuff_dbug (char *srcname, double ipkttime, char *packet, int nbytes, Packet * pkt)
{
    char           *pp,
                   *cp;
    int             retcode = 0;
    int             dummy = 0;
    int             ICHAN = 0;
    PktChannel     *achan;

    achan = gettbl (pkt->channels, 0);
    if (achan == 0) {
	achan = newPktChannel ();
	pushtbl (pkt->channels, achan);
    }
    pp = cp = packet;
    NC2HI (&pkt->version, cp, 1);
    cp += 1 * 1;
    switch (pkt->version) {
      case 0:
	cp += 7;
	break;
    }
    achan->time = ipkttime;
    N2H4 ((char *) &pkt->nchannels, cp, 1);
    cp += 4 * 1;
    if (pkt->nchannels > 12)
	return -1;
    for (ICHAN = 0; ICHAN < pkt->nchannels; ICHAN++) {
	achan = gettbl (pkt->channels, ICHAN);
	if (achan == 0) {
	    achan = newPktChannel ();
	    pushtbl (pkt->channels, achan);
	};
	memcpy (&achan->net, cp, PKT_NAMESIZE);
	cp += 1 * PKT_NAMESIZE;
	memcpy (&achan->sta, cp, PKT_NAMESIZE);
	cp += 1 * PKT_NAMESIZE;
	memcpy (&achan->chan, cp, PKT_NAMESIZE);
	cp += 1 * PKT_NAMESIZE;
	N2HD (&achan->time, (uchar *) cp, 1);
	cp += 8 * 1;
	N2HD (&achan->samprate, (uchar *) cp, 1);
	cp += 8 * 1;
	N2HD (&achan->calib, (uchar *) cp, 1);
	cp += 8 * 1;
	switch (pkt->version) {
	  case 0:
	    N2H4 ((char *) &achan->nsamp, cp, 1);
	    cp += 4 * 1;
	    N2H4 ((char *) &dummy, cp, 1);
	    cp += 4 * 1;
	    N2H4 ((char *) &dummy, cp, 1);
	    cp += 4 * 1;
	    break;
	  default:
	    N2HD (&achan->calper, (uchar *) cp, 1);
	    cp += 8 * 1;
	    memcpy (&achan->segtype, cp, 2);
	    cp += 1 * 2;
	    N2H4 ((char *) &achan->nsamp, cp, 1);
	    cp += 4 * 1;
	    break;
	}
    };
    for (ICHAN = 0; ICHAN < pkt->nchannels; ICHAN++) {
	achan = gettbl (pkt->channels, ICHAN);
	if (achan == 0) {
	    achan = newPktChannel ();
	    pushtbl (pkt->channels, achan);
	};
	RESIZE_BUFFER (int *, achan->data, achan->datasz, achan->nsamp + 1);
	N2H4 ((char *) achan->data, cp, achan->nsamp);
	cp += 4 * achan->nsamp;
    };
    if (nbytes < cp - pp) {
	complain (0, "nbytes=%d < cp-pp=%d\n", nbytes, cp - pp);
	retcode = -1;
    };
    return retcode;
}


int
stuff_Cougar (Packet * pkt, char *srcname, double *opkttime, char **ppp, int *nbytes, int *ppsz)
{
    char           *pp,
                   *cp;
    int             retcode = 0;
    short           hdrsize = 30 + 24 + 12 + 12;
    short           datasize;
    char           *pdatasize;
    int             datatype = 3;
    int             databytes;
    short           type = 0xdeb;
    int             ICHAN = 0;
    PktChannel     *achan;
    achan = gettbl (pkt->channels, 0);
    pp = cp = *ppp;
    RESIZE_BUFFER (char *, *ppp, *ppsz, (20 + (36 + 3 * PKT_NAMESIZE + achan->nsamp * 4) * pkt->nchannels));
    cp = *ppp + (cp - pp);
    pp = *ppp;
    *opkttime = achan->time;
    H2N2 (cp, (char *) &hdrsize, 1);
    cp += 2 * 1;
    pdatasize = cp;
    H2N2 (cp, (char *) &datasize, 1);
    cp += 2 * 1;
    H2N2 (cp, (char *) &type, 1);
    cp += 2 * 1;
    H2N2 (cp, (char *) &type, 1);
    cp += 2 * 1;
    H2N4 (cp, (char *) &pkt->nchannels, 1);
    cp += 4 * 1;
    memcpy (cp, &achan->net, PKT_NAMESIZE);
    cp += 1 * PKT_NAMESIZE;
    memcpy (cp, &achan->sta, PKT_NAMESIZE);
    cp += 1 * PKT_NAMESIZE;
    memcpy (cp, &achan->chan, PKT_NAMESIZE);
    cp += 1 * PKT_NAMESIZE;
    H2ND ((uchar *) cp, &achan->time, 1);
    cp += 8 * 1;
    H2ND ((uchar *) cp, &achan->samprate, 1);
    cp += 8 * 1;
    H2ND ((uchar *) cp, &achan->calib, 1);
    cp += 8 * 1;
    H2N4 (cp, (char *) &achan->nsamp, 1);
    cp += 4 * 1;
    databytes = 4 * achan->nsamp;
    H2N4 (cp, (char *) &databytes, 1);
    cp += 4 * 1;
    H2N4 (cp, (char *) &datatype, 1);
    cp += 4 * 1;
    H2N4 (cp, (char *) achan->data, achan->nsamp);
    cp += 4 * achan->nsamp;
    datasize = 4 * achan->nsamp;
    H2N2 (pdatasize, (char *) &datasize, 1);
    *nbytes = cp - pp;
    insist (*nbytes < (20 + (36 + 3 * PKT_NAMESIZE + achan->nsamp * 4) * pkt->nchannels));
    return retcode;
}



int
unstuff_Cougar (char *srcname, double ipkttime, char *packet, int nbytes, Packet * pkt)
{
    char           *pp,
                   *cp;
    int             retcode = 0;
    short           hdrsize = 30 + 24 + 12 + 12;
    short           datasize;
    int             datatype = 3;
    int             databytes;
    short           type = 0xdeb;
    int             ICHAN = 0;
    PktChannel     *achan;

    achan = gettbl (pkt->channels, 0);
    if (achan == 0) {
	achan = newPktChannel ();
	pushtbl (pkt->channels, achan);
    }
    pp = cp = packet;
    achan->time = ipkttime;
    N2H2 ((char *) &hdrsize, cp, 1);
    cp += 2 * 1;
    if (hdrsize > nbytes)
	return -1;
    N2H2 ((char *) &datasize, cp, 1);
    cp += 2 * 1;
    if (datasize + hdrsize > nbytes)
	return -1;
    N2H2 ((char *) &type, cp, 1);
    cp += 2 * 1;
    N2H2 ((char *) &type, cp, 1);
    cp += 2 * 1;
    N2H4 ((char *) &pkt->nchannels, cp, 1);
    cp += 4 * 1;
    if (pkt->nchannels > 1)
	return -1;
    memcpy (&achan->net, cp, PKT_NAMESIZE);
    cp += 1 * PKT_NAMESIZE;
    memcpy (&achan->sta, cp, PKT_NAMESIZE);
    cp += 1 * PKT_NAMESIZE;
    memcpy (&achan->chan, cp, PKT_NAMESIZE);
    cp += 1 * PKT_NAMESIZE;
    N2HD (&achan->time, (uchar *) cp, 1);
    cp += 8 * 1;
    N2HD (&achan->samprate, (uchar *) cp, 1);
    cp += 8 * 1;
    N2HD (&achan->calib, (uchar *) cp, 1);
    cp += 8 * 1;
    N2H4 ((char *) &achan->nsamp, cp, 1);
    cp += 4 * 1;
    if (achan->nsamp < 0 || achan->nsamp > MAX_NSAMP)
	return -1;
    N2H4 ((char *) &databytes, cp, 1);
    cp += 4 * 1;
    N2H4 ((char *) &datatype, cp, 1);
    cp += 4 * 1;
    RESIZE_BUFFER (int *, achan->data, achan->datasz, achan->nsamp + 1);
    N2H4 ((char *) achan->data, cp, achan->nsamp);
    cp += 4 * achan->nsamp;
    if (nbytes < cp - pp) {
	complain (0, "nbytes=%d < cp-pp=%d\n", nbytes, cp - pp);
	retcode = -1;
    };
    return retcode;
}


int
stuff_iceworm (Packet * pkt, char *srcname, double *opkttime, char **ppp, int *nbytes, int *ppsz)
{
    char           *pp,
                   *cp;
    int             retcode = 0;
    unsigned char  *buf = 0;
    int             bufsz = 0;
    int             bytecount;
    int             i;
    int            *idata;
    int             ICHAN = 0;
    PktChannel     *achan;
    achan = gettbl (pkt->channels, 0);
    pp = cp = *ppp;
    RESIZE_BUFFER (char *, *ppp, *ppsz, 48 + 5 * achan->nsamp);
    cp = *ppp + (cp - pp);
    pp = *ppp;
    H2N2 (cp, (char *) &Zero, 1);
    cp += 2 * 1;
    H2N2 (cp, (char *) &Zero, 1);
    cp += 2 * 1;
    HI2NS (cp, (int *) &pkt->pkttype->hdrcode, 1);
    cp += 2 * 1;
    HI2NS (cp, (int *) &pkt->pkttype->bodycode, 1);
    cp += 2 * 1;
    HD2NF (cp, &achan->samprate, 1);
    cp += 4 * 1;
    HD2NF (cp, &achan->calib, 1);
    cp += 4 * 1;
    HI2NS (cp, (int *) &achan->iuser1, 1);
    cp += 2 * 1;
    HI2NS (cp, (int *) &achan->nsamp, 1);
    cp += 2 * 1;
    memcpy (cp, achan->cuser1, 2);
    cp += 1 * 2;
    HI2NS (cp, (int *) &achan->iuser2, 1);
    cp += 2 * 1;
    *opkttime = achan->time;
    bytecount = cp - pp;
    HI2NS (pp, &bytecount, 1);
    if (strcmp (achan->cuser1, "s4") == 0) {
	H2N4 (cp, (char *) achan->data, achan->nsamp);
	cp += achan->nsamp * sizeof (int);
    } else if (strcmp (achan->cuser1, "s2") == 0) {
	HI2NS (cp, achan->data, achan->nsamp);
	cp += achan->nsamp * sizeof (short);
    } else if (strcmp (achan->cuser1, "t4") == 0) {
	HI2NF (cp, achan->data, achan->nsamp);
	cp += achan->nsamp * sizeof (float);
    } else if (strcmp (achan->cuser1, "gc") == 0 || achan->cuser1[0] == 0) {
	allot (int *, idata, achan->nsamp);
	idata[0] = achan->data[0];
	for (i = 1; i < achan->nsamp; i++)
	    idata[i] = achan->data[i] - achan->data[i - 1];
	retcode = gencompress ((unsigned char **) &buf, &bytecount, &bufsz, idata, achan->nsamp, 0);
	free (idata);
	memcpy (cp, buf, bytecount);
	free (buf);
	cp += bytecount;
    } else {
	retcode = -1;
    };
    bytecount = cp - pp;
    HI2NS (pp + 2, &bytecount, 1);
    *nbytes = cp - pp;
    insist (*nbytes < 48 + 5 * achan->nsamp);
    return retcode;
}



int
unstuff_iceworm (char *srcname, double ipkttime, char *packet, int nbytes, Packet * pkt)
{
    char           *pp,
                   *cp;
    int             retcode = 0;
    unsigned char  *buf = 0;
    int             bufsz = 0;
    int             i;
    char            datatype[4];
    int             ICHAN = 0;
    PktChannel     *achan;

    achan = gettbl (pkt->channels, 0);
    if (achan == 0) {
	achan = newPktChannel ();
	pushtbl (pkt->channels, achan);
    }
    pp = cp = packet;
    cp += 8;
    pkt->nchannels = 1;
    strncpy (achan->segtype, "V", 4);
    NF2HD (&achan->samprate, cp, 1);
    cp += 4 * 1;
    NF2HD (&achan->calib, cp, 1);
    cp += 4 * 1;
    NS2HI (&achan->iuser1, cp, 1);
    cp += 2 * 1;
    NS2HI (&achan->nsamp, cp, 1);
    cp += 2 * 1;
    if (achan->nsamp < 0 || achan->nsamp > MAX_NSAMP)
	return -1;
    memcpy (achan->cuser1, cp, 2);
    cp += 1 * 2;
    achan->cuser1[2] = 0;
    NS2HI (&achan->iuser2, cp, 1);
    cp += 2 * 1;
    strncpy (achan->net, pkt->parts.src_net, PKT_NAMESIZE);
    strncpy (achan->sta, pkt->parts.src_sta, PKT_NAMESIZE);
    strncpy (achan->chan, pkt->parts.src_chan, PKT_NAMESIZE);
    strncpy (achan->loc, pkt->parts.src_loc, PKT_NAMESIZE);
    achan->time = ipkttime;
    RESIZE_BUFFER (int *, achan->data, achan->datasz, achan->nsamp + 1);
    if (strcmp (achan->cuser1, "s4") == 0) {
	N2H4 ((char *) achan->data, cp, achan->nsamp);
	cp += achan->nsamp * sizeof (int);
    } else if (strcmp (achan->cuser1, "s2") == 0) {
	NS2HI (achan->data, cp, achan->nsamp);
	cp += achan->nsamp * sizeof (short);
    } else if (strcmp (achan->cuser1, "t4") == 0) {
	NF2HI (achan->data, cp, achan->nsamp);
	cp += achan->nsamp * sizeof (float);
    } else if (strcmp (achan->cuser1, "gc") == 0 || achan->cuser1[0] == 0) {
	achan->datasz *= sizeof (int);
	retcode = genuncompress (&achan->data, &achan->nsamp, &achan->datasz, (unsigned char *) cp, nbytes - (cp - pp));
	achan->datasz /= sizeof (int);
	for (i = 1; i < achan->nsamp; i++)
	    achan->data[i] += achan->data[i - 1];
    } else {
	retcode = -1;
    };
    if (nbytes < cp - pp) {
	complain (0, "nbytes=%d < cp-pp=%d\n", nbytes, cp - pp);
	retcode = -1;
    };
    return retcode;
}


int
stuff_replay (Packet * pkt, char *srcname, double *opkttime, char **ppp, int *nbytes, int *ppsz)
{
    char           *pp,
                   *cp;
    int             retcode = 0;
    unsigned char  *buf = 0;
    int             bufsz = 0;
    int             bytecount;
    int             i;
    int            *idata;
    int             ICHAN = 0;
    PktChannel     *achan;
    achan = gettbl (pkt->channels, 0);
    pp = cp = *ppp;
    RESIZE_BUFFER (char *, *ppp, *ppsz, 64 + 5 * achan->nsamp);
    cp = *ppp + (cp - pp);
    pp = *ppp;
    pkt->version = 1;
    HI2NC (cp, &pkt->version, 1);
    cp += 1 * 1;
    switch (pkt->version) {
      case 0:
	memcpy (cp, &Zero, 1);
	cp += 1 * 1;
	H2N2 (cp, (char *) &Zero, 1);
	cp += 2 * 1;
	HI2NS (cp, (int *) &pkt->pkttype->hdrcode, 1);
	cp += 2 * 1;
	HI2NS (cp, (int *) &pkt->pkttype->bodycode, 1);
	cp += 2 * 1;
	break;
      case 1:
	HD2NF (cp, &achan->calper, 1);
	cp += 4 * 1;
	memcpy (cp, &achan->segtype, 2);
	cp += 1 * 2;
	break;
      default:
	return -1;
    }
    HD2NF (cp, &achan->samprate, 1);
    cp += 4 * 1;
    HD2NF (cp, &achan->calib, 1);
    cp += 4 * 1;
    HI2NS (cp, (int *) &achan->iuser1, 1);
    cp += 2 * 1;
    HI2NS (cp, (int *) &achan->nsamp, 1);
    cp += 2 * 1;
    memcpy (cp, achan->cuser1, 2);
    cp += 1 * 2;
    HI2NS (cp, (int *) &achan->iuser2, 1);
    cp += 2 * 1;
    *opkttime = achan->time;
    if (strcmp (achan->cuser1, "s4") == 0) {
	H2N4 (cp, (char *) achan->data, achan->nsamp);
	cp += achan->nsamp * sizeof (int);
    } else if (strcmp (achan->cuser1, "s2") == 0) {
	HI2NS (cp, achan->data, achan->nsamp);
	cp += achan->nsamp * sizeof (short);
    } else if (strcmp (achan->cuser1, "t4") == 0) {
	HI2NF (cp, achan->data, achan->nsamp);
	cp += achan->nsamp * sizeof (short);
    } else if (strcmp (achan->cuser1, "gc") == 0 || achan->cuser1[0] == 0) {
	allot (int *, idata, achan->nsamp);
	idata[0] = achan->data[0];
	for (i = 1; i < achan->nsamp; i++)
	    idata[i] = achan->data[i] - achan->data[i - 1];
	retcode = gencompress ((unsigned char **) &buf, &bytecount, &bufsz, idata, achan->nsamp, 0);
	free (idata);
	memcpy (cp, buf, bytecount);
	free (buf);
	cp += bytecount;
    } else {
	retcode = -1;
    };
    *nbytes = cp - pp;
    insist (*nbytes < 64 + 5 * achan->nsamp);
    return retcode;
}



int
unstuff_replay (char *srcname, double ipkttime, char *packet, int nbytes, Packet * pkt)
{
    char           *pp,
                   *cp;
    int             retcode = 0;
    unsigned char  *buf = 0;
    int             bufsz = 0;
    int             i;
    int             ICHAN = 0;
    PktChannel     *achan;

    achan = gettbl (pkt->channels, 0);
    if (achan == 0) {
	achan = newPktChannel ();
	pushtbl (pkt->channels, achan);
    }
    pp = cp = packet;
    NC2HI (&pkt->version, cp, 1);
    cp += 1 * 1;
    switch (pkt->version) {
      case 0:
	cp += 7;
	strncpy (achan->segtype, "V", 4);
	break;
      case 1:
	NF2HD (&achan->calper, cp, 1);
	cp += 4 * 1;
	memcpy (&achan->segtype, cp, 2);
	cp += 1 * 2;
	break;
      default:
	return -1;
    }
    pkt->nchannels = 1;
    NF2HD (&achan->samprate, cp, 1);
    cp += 4 * 1;
    NF2HD (&achan->calib, cp, 1);
    cp += 4 * 1;
    NS2HI (&achan->iuser1, cp, 1);
    cp += 2 * 1;
    NS2HI (&achan->nsamp, cp, 1);
    cp += 2 * 1;
    if (achan->nsamp < 0 || achan->nsamp > 60000)
	return -1;
    memcpy (achan->cuser1, cp, 2);
    cp += 1 * 2;
    achan->cuser1[2] = 0;
    NS2HI (&achan->iuser2, cp, 1);
    cp += 2 * 1;
    strncpy (achan->net, pkt->parts.src_net, PKT_NAMESIZE);
    strncpy (achan->sta, pkt->parts.src_sta, PKT_NAMESIZE);
    strncpy (achan->chan, pkt->parts.src_chan, PKT_NAMESIZE);
    strncpy (achan->loc, pkt->parts.src_loc, PKT_NAMESIZE);
    achan->time = ipkttime;
    RESIZE_BUFFER (int *, achan->data, achan->datasz, achan->nsamp + 1);
    if (strcmp (achan->cuser1, "s4") == 0) {
	N2H4 ((char *) achan->data, cp, achan->nsamp);
	cp += achan->nsamp * sizeof (int);
    } else if (strcmp (achan->cuser1, "s2") == 0) {
	NS2HI (achan->data, cp, achan->nsamp);
	cp += achan->nsamp * sizeof (short);
    } else if (strcmp (achan->cuser1, "t4") == 0) {
	NF2HI (achan->data, cp, achan->nsamp);
	cp += achan->nsamp * sizeof (short);
    } else if (strcmp (achan->cuser1, "gc") == 0 || achan->cuser1[0] == 0) {
	achan->datasz *= sizeof (int);
	retcode = genuncompress (&achan->data, &achan->nsamp, &achan->datasz, (unsigned char *) cp, nbytes - (cp - pp));
	achan->datasz /= sizeof (int);
	for (i = 1; i < achan->nsamp; i++)
	    achan->data[i] += achan->data[i - 1];
    } else {
	retcode = -1;
    };
    if (nbytes < cp - pp) {
	complain (0, "nbytes=%d < cp-pp=%d\n", nbytes, cp - pp);
	retcode = -1;
    };
    return retcode;
}


int
stuff_CPSCLHS (Packet * pkt, char *srcname, double *opkttime, char **ppp, int *nbytes, int *ppsz)
{
    char           *pp,
                   *cp;
    int             retcode = 0;
    double          lcalib;
    float           lsamprate;
    int             lnsamp,
                    ldatatype;
    int             doff;
    short           chlen;
    char            channel_names[ORBSRCNAME_SIZE],
                   *aname;
    char           *lasts;
    int             ICHAN = 0;
    PktChannel     *achan;
    achan = gettbl (pkt->channels, 0);
    pp = cp = *ppp;
    RESIZE_BUFFER (char *, *ppp, *ppsz, (ORBSRCNAME_SIZE + 20 + (achan->nsamp * 4 + 12) * pkt->nchannels));
    cp = *ppp + (cp - pp);
    pp = *ppp;
    H2N2 (cp, (char *) &Zero, 1);
    cp += 2 * 1;
    H2N2 (cp, (char *) &Zero, 1);
    cp += 2 * 1;
    HI2NS (cp, (int *) &pkt->pkttype->hdrcode, 1);
    cp += 2 * 1;
    HI2NS (cp, (int *) &pkt->pkttype->bodycode, 1);
    cp += 2 * 1;
    *opkttime = achan->time;
    HD2NF (cp, &achan->calib, 1);
    cp += 4 * 1;
    HD2NF (cp, &achan->samprate, 1);
    cp += 4 * 1;
    HI2NS (cp, (int *) &achan->iuser1, 1);
    cp += 2 * 1;
    HI2NS (cp, (int *) &achan->nsamp, 1);
    cp += 2 * 1;
    HI2NS (cp, (int *) &pkt->nchannels, 1);
    cp += 2 * 1;
    strcpy (channel_names, "_");
    for (ICHAN = 0; ICHAN < pkt->nchannels; ICHAN++) {
	achan = gettbl (pkt->channels, ICHAN);
	if (achan == 0) {
	    achan = newPktChannel ();
	    pushtbl (pkt->channels, achan);
	};
	strcat (channel_names, "_");
	strcat (channel_names, achan->chan);
    };
    strncpy ((char *) cp, (char *) channel_names, ORBSRCNAME_SIZE);
    cp += MIN (strlen (channel_names) + 1, ORBSRCNAME_SIZE);
    for (ICHAN = 0; ICHAN < pkt->nchannels; ICHAN++) {
	achan = gettbl (pkt->channels, ICHAN);
	if (achan == 0) {
	    achan = newPktChannel ();
	    pushtbl (pkt->channels, achan);
	};
	switch (achan->iuser1) {
	  case trINT:
	    H2N4 (cp, (char *) achan->data, achan->nsamp);
	    cp += 4 * achan->nsamp;
	    break;
	  case trSHORT:
	    HI2NS (cp, (int *) achan->data, achan->nsamp);
	    cp += 2 * achan->nsamp;
	    break;
	  case 0:
	    retcode = -1;
	    break;
	}
    };
    *nbytes = cp - pp;
    insist (*nbytes < (ORBSRCNAME_SIZE + 20 + (achan->nsamp * 4 + 12) * pkt->nchannels));
    return retcode;
}



int
unstuff_CPSCLHS (char *srcname, double ipkttime, char *packet, int nbytes, Packet * pkt)
{
    char           *pp,
                   *cp;
    int             retcode = 0;
    double          lcalib;
    float           lsamprate;
    int             lnsamp,
                    ldatatype;
    int             doff;
    short           chlen;
    char            channel_names[ORBSRCNAME_SIZE],
                   *aname;
    char           *lasts;
    int             ICHAN = 0;
    PktChannel     *achan;

    achan = gettbl (pkt->channels, 0);
    if (achan == 0) {
	achan = newPktChannel ();
	pushtbl (pkt->channels, achan);
    }
    pp = cp = packet;
    cp += 4;
    NS2HI (&pkt->version, cp, 1);
    cp += 2 * 1;
    cp += 2;
    switch (pkt->version) {
      case 0xddd:
	cp += PKT_NAMESIZE;
	cp += PKT_NAMESIZE;
	strncpy (channel_names, cp, ORBSRCNAME_SIZE);
	cp += ORBSRCNAME_SIZE;
	N2H4 ((char *) &ldatatype, cp, 1);
	cp += 4 * 1;
	cp += 8;
	N2HD (&lcalib, (uchar *) cp, 1);
	cp += 8 * 1;
	N2H4 ((char *) &lsamprate, cp, 1);
	cp += 4 * 1;
	N2H4 ((char *) &lnsamp, cp, 1);
	cp += 4 * 1;
	N2H4 ((char *) &pkt->nchannels, cp, 1);
	cp += 4 * 1;
	N2H4 ((char *) &doff, cp, 1);
	cp += 4 * 1;
	break;
      default:
	return -1;
	break;
      case 0xbba:
	NF2HD (&lcalib, cp, 1);
	cp += 4 * 1;
	N2H4 ((char *) &lsamprate, cp, 1);
	cp += 4 * 1;
	NS2HI (&ldatatype, cp, 1);
	cp += 2 * 1;
	NS2HI (&lnsamp, cp, 1);
	cp += 2 * 1;
	if (lnsamp < 0 || lnsamp > MAX_NSAMP) {
	    complain (0, "achan->nsamp=%d", achan->nsamp);
	    return -1;
	};
	NS2HI (&pkt->nchannels, cp, 1);
	cp += 2 * 1;
	NS2HI (&doff, cp, 1);
	cp += 2 * 1;
	N2H2 ((char *) &chlen, cp, 1);
	cp += 2 * 1;
	if (chlen <= 0 || chlen > ORBSRCNAME_SIZE)
	    return -1;
	memcpy (&channel_names, cp, chlen);
	cp += 1 * chlen;
	channel_names[chlen] = 0;
	break;
    };
    if (pkt->nchannels > 3 || pkt->nchannels < 1) {
	complain (0, "pkt->nchannels=%d", pkt->nchannels);
	return -1;
    };
    aname = strtok_r (channel_names, "_", &lasts);
    cp += doff;
    for (ICHAN = 0; ICHAN < pkt->nchannels; ICHAN++) {
	achan = gettbl (pkt->channels, ICHAN);
	if (achan == 0) {
	    achan = newPktChannel ();
	    pushtbl (pkt->channels, achan);
	};
	strncpy (achan->net, pkt->parts.src_net, PKT_NAMESIZE);
	strncpy (achan->sta, pkt->parts.src_sta, PKT_NAMESIZE);
	if (aname == 0)
	    return -1;
	strcpy (achan->chan, aname);
	achan->time = ipkttime;
	aname = strtok_r (0, "_", &lasts);
	achan->samprate = lsamprate;
	achan->calib = lcalib;
	strcpy (achan->segtype, "V");
	achan->calper = 0.0;
	achan->nsamp = lnsamp;
	RESIZE_BUFFER (int *, achan->data, achan->datasz, achan->nsamp + 1);
	achan->iuser1 = ldatatype;
	achan->puser1 = 0;
	achan->iuser2 = 0;
	switch (achan->iuser1) {
	  case trINT:
	    N2H4 ((char *) achan->data, cp, achan->nsamp);
	    cp += 4 * achan->nsamp;
	    break;
	  case trSHORT:
	    NS2HI (achan->data, cp, achan->nsamp);
	    cp += 2 * achan->nsamp;
	    break;
	  case 0:
	    achan->puser1 = cp - 64;
	    achan->iuser2 = nbytes - (cp - pp) + 64;
	    if (achan->iuser1 != 0)
		retcode = -1;
	    else
		retcode += upscl (achan->puser1, achan->nsamp, achan->data);
	    break;
	}
    };
    if (nbytes < cp - pp) {
	complain (0, "nbytes=%d < cp-pp=%d\n", nbytes, cp - pp);
	retcode = -1;
    };
    return retcode;
}


int
stuff_CPSCLLS (Packet * pkt, char *srcname, double *opkttime, char **ppp, int *nbytes, int *ppsz)
{
    return stuff_CPSCLHS (pkt, srcname, opkttime, ppp, nbytes, ppsz);
}

int
unstuff_CPSCLLS (char *srcname, double ipkttime, char *packet, int nbytes, Packet * pkt)
{
    return unstuff_CPSCLHS (srcname, ipkttime, packet, nbytes, pkt);
}


int
stuff_BBA (Packet * pkt, char *srcname, double *opkttime, char **ppp, int *nbytes, int *ppsz)
{
    char           *pp,
                   *cp;
    int             retcode = 0;
    int             i;
    double          lcalib;
    float           lsamprate;
    int             lnsamp,
                    ldatatype;
    int             doff,
                    choff,
                    chbytes;
    short           chlen;
    unsigned short  val;
    char            channel_names[ORBSRCNAME_SIZE * 2],
                   *aname;
    char           *lasts;
    int             ICHAN = 0;
    PktChannel     *achan;
    achan = gettbl (pkt->channels, 0);
    pp = cp = *ppp;
    RESIZE_BUFFER (char *, *ppp, *ppsz, (ORBSRCNAME_SIZE + 20 + (achan->nsamp * 4 + 12) * pkt->nchannels));
    cp = *ppp + (cp - pp);
    pp = *ppp;
    H2N2 (cp, (char *) &Zero, 1);
    cp += 2 * 1;
    H2N2 (cp, (char *) &Zero, 1);
    cp += 2 * 1;
    HI2NS (cp, (int *) &pkt->pkttype->hdrcode, 1);
    cp += 2 * 1;
    HI2NS (cp, (int *) &pkt->pkttype->bodycode, 1);
    cp += 2 * 1;
    *opkttime = achan->time;
    HD2NF (cp, &achan->calib, 1);
    cp += 4 * 1;
    HD2NF (cp, &achan->samprate, 1);
    cp += 4 * 1;
    HI2NS (cp, (int *) &achan->iuser1, 1);
    cp += 2 * 1;
    HI2NS (cp, (int *) &achan->nsamp, 1);
    cp += 2 * 1;
    HI2NS (cp, (int *) &pkt->nchannels, 1);
    cp += 2 * 1;
    strcpy (channel_names, "_");
    for (ICHAN = 0; ICHAN < pkt->nchannels; ICHAN++) {
	achan = gettbl (pkt->channels, ICHAN);
	if (achan == 0) {
	    achan = newPktChannel ();
	    pushtbl (pkt->channels, achan);
	};
	strcat (channel_names, "_");
	strcat (channel_names, achan->chan);
    };
    strncpy ((char *) cp, (char *) channel_names, ORBSRCNAME_SIZE * 2);
    cp += MIN (strlen (channel_names) + 1, ORBSRCNAME_SIZE * 2);
    for (ICHAN = 0; ICHAN < pkt->nchannels; ICHAN++) {
	achan = gettbl (pkt->channels, ICHAN);
	if (achan == 0) {
	    achan = newPktChannel ();
	    pushtbl (pkt->channels, achan);
	};
	switch (achan->iuser1) {
	  case 3:		       /* integer */
	    H2N4 (cp, (char *) achan->data, achan->nsamp);
	    cp += 4 * achan->nsamp;
	    break;
	  case 2:		       /* short */
	    HI2NS (cp, (int *) achan->data, achan->nsamp);
	    cp += 2 * achan->nsamp;
	    break;
	  case 0:
	  default:
	    retcode = -1;
	    break;
	}
    };
    *nbytes = cp - pp;
    insist (*nbytes < (ORBSRCNAME_SIZE + 20 + (achan->nsamp * 4 + 12) * pkt->nchannels));
    return retcode;
}



int
unstuff_BBA (char *srcname, double ipkttime, char *packet, int nbytes, Packet * pkt)
{
    char           *pp,
                   *cp;
    int             retcode = 0;
    int             i;
    double          lcalib;
    float           lsamprate;
    int             lnsamp,
                    ldatatype;
    int             doff,
                    choff,
                    chbytes;
    short           chlen;
    unsigned short  val;
    char            channel_names[ORBSRCNAME_SIZE * 2],
                   *aname;
    char           *lasts;
    int             ICHAN = 0;
    PktChannel     *achan;

    achan = gettbl (pkt->channels, 0);
    if (achan == 0) {
	achan = newPktChannel ();
	pushtbl (pkt->channels, achan);
    }
    pp = cp = packet;
    cp += 4;
    NS2HI (&pkt->version, cp, 1);
    cp += 2 * 1;
    cp += 2;
    switch (pkt->version) {
      case 0xddd:
	cp += PKT_NAMESIZE;
	cp += PKT_NAMESIZE;
	strncpy (channel_names, cp, ORBSRCNAME_SIZE);
	cp += ORBSRCNAME_SIZE;
	N2H4 ((char *) &ldatatype, cp, 1);
	cp += 4 * 1;
	cp += 8;
	N2HD (&lcalib, (uchar *) cp, 1);
	cp += 8 * 1;
	N2H4 ((char *) &lsamprate, cp, 1);
	cp += 4 * 1;
	N2H4 ((char *) &lnsamp, cp, 1);
	cp += 4 * 1;
	N2H4 ((char *) &pkt->nchannels, cp, 1);
	cp += 4 * 1;
	N2H4 ((char *) &doff, cp, 1);
	cp += 4 * 1;
	break;
      default:
	return -1;
	break;
      case 0xbba:
	NF2HD (&lcalib, cp, 1);
	cp += 4 * 1;
	N2H4 ((char *) &lsamprate, cp, 1);
	cp += 4 * 1;
	NS2HI (&ldatatype, cp, 1);
	cp += 2 * 1;
	NS2HI (&lnsamp, cp, 1);
	cp += 2 * 1;
	NS2HI (&pkt->nchannels, cp, 1);
	cp += 2 * 1;
	NS2HI (&doff, cp, 1);
	cp += 2 * 1;
	N2H2 ((char *) &chlen, cp, 1);
	cp += 2 * 1;
	if (chlen <= 0)
	    return -1;
	memcpy (&channel_names, cp, chlen);
	cp += 1 * chlen;
	channel_names[chlen] = 0;
	break;
    };
    if (lnsamp < 1 || lnsamp > MAX_NSAMP) {
	complain (0, "lnsamp=%d", lnsamp);
	return -1;
    };
    if (pkt->nchannels < 1) {
	complain (0, "pkt->nchannels=%d", pkt->nchannels);
	return -1;
    };
    aname = strtok_r (channel_names, "_", &lasts);
    cp += doff;
    for (ICHAN = 0; ICHAN < pkt->nchannels; ICHAN++) {
	achan = gettbl (pkt->channels, ICHAN);
	if (achan == 0) {
	    achan = newPktChannel ();
	    pushtbl (pkt->channels, achan);
	};
	strncpy (achan->net, pkt->parts.src_net, PKT_NAMESIZE);
	strncpy (achan->sta, pkt->parts.src_sta, PKT_NAMESIZE);
	if (aname == 0)
	    return -1;
	strcpy (achan->chan, aname);
	achan->time = ipkttime;
	aname = strtok_r (0, "_", &lasts);
	achan->samprate = lsamprate;
	achan->calib = lcalib;
	strcpy (achan->segtype, "V");
	achan->calper = 0.0;
	achan->nsamp = lnsamp;
	if (achan->nsamp < 0 || achan->nsamp > MAX_NSAMP)
	    return -1;
	RESIZE_BUFFER (int *, achan->data, achan->datasz, achan->nsamp + 1);
	achan->iuser1 = ldatatype;
	switch (achan->iuser1) {
	  case 3:		       /* integer */
	    cp += 4;
	    N2H4 ((char *) achan->data, cp, achan->nsamp);
	    cp += 4 * achan->nsamp;
	    break;
	  case 2:		       /* short */
	    cp += 4;
	    NS2HI (achan->data, cp, achan->nsamp);
	    cp += 2 * achan->nsamp;
	    break;
	  case 0:
	  default:
	    val = ((unsigned char) cp[2]) * 256 + ((unsigned char) cp[3]);
	    chbytes = val;
	    cp += 4;
	    if (achan->iuser1 != 0)
		retcode = -1;
	    else
		retcode += uucsd (achan->data, &cp, chbytes, achan->nsamp);
	    break;
	}
    };
    if (nbytes < cp - pp) {
	complain (0, "nbytes=%d < cp-pp=%d\n", nbytes, cp - pp);
	retcode = -1;
    };
    return retcode;
}


int
stuff_CBBHS (Packet * pkt, char *srcname, double *opkttime, char **ppp, int *nbytes, int *ppsz)
{
    char           *pp,
                   *cp;
    int             retcode = 0;
    int             i;
    double          lcalib;
    float           lsamprate;
    int             lnsamp,
                    ldatatype;
    int             doff;
    short           chlen;
    char            channel_names[ORBSRCNAME_SIZE],
                   *aname;
    char           *lasts;
    int             ICHAN = 0;
    PktChannel     *achan;
    achan = gettbl (pkt->channels, 0);
    pp = cp = *ppp;
    RESIZE_BUFFER (char *, *ppp, *ppsz, (ORBSRCNAME_SIZE + 20 + (achan->nsamp * 4 + 12) * pkt->nchannels));
    cp = *ppp + (cp - pp);
    pp = *ppp;
    H2N2 (cp, (char *) &Zero, 1);
    cp += 2 * 1;
    H2N2 (cp, (char *) &Zero, 1);
    cp += 2 * 1;
    HI2NS (cp, (int *) &pkt->pkttype->hdrcode, 1);
    cp += 2 * 1;
    HI2NS (cp, (int *) &pkt->pkttype->bodycode, 1);
    cp += 2 * 1;
    *opkttime = achan->time;
    HD2NF (cp, &achan->calib, 1);
    cp += 4 * 1;
    HD2NF (cp, &achan->samprate, 1);
    cp += 4 * 1;
    HI2NS (cp, (int *) &achan->iuser1, 1);
    cp += 2 * 1;
    HI2NS (cp, (int *) &achan->nsamp, 1);
    cp += 2 * 1;
    HI2NS (cp, (int *) &pkt->nchannels, 1);
    cp += 2 * 1;
    strcpy (channel_names, "_");
    for (ICHAN = 0; ICHAN < pkt->nchannels; ICHAN++) {
	achan = gettbl (pkt->channels, ICHAN);
	if (achan == 0) {
	    achan = newPktChannel ();
	    pushtbl (pkt->channels, achan);
	};
	strcat (channel_names, "_");
	strcat (channel_names, achan->chan);
    };
    strncpy ((char *) cp, (char *) channel_names, ORBSRCNAME_SIZE);
    cp += MIN (strlen (channel_names) + 1, ORBSRCNAME_SIZE);
    for (ICHAN = 0; ICHAN < pkt->nchannels; ICHAN++) {
	achan = gettbl (pkt->channels, ICHAN);
	if (achan == 0) {
	    achan = newPktChannel ();
	    pushtbl (pkt->channels, achan);
	};
	switch (achan->iuser1) {
	  case 3:		       /* integer */
	    H2N4 (cp, (char *) achan->data, achan->nsamp);
	    cp += 4 * achan->nsamp;
	    break;
	  case 2:		       /* short */
	    HI2NS (cp, (int *) achan->data, achan->nsamp);
	    cp += 2 * achan->nsamp;
	    break;
	  case 0:
	  default:
	    retcode = -1;
	    break;
	}
    };
    *nbytes = cp - pp;
    insist (*nbytes < (ORBSRCNAME_SIZE + 20 + (achan->nsamp * 4 + 12) * pkt->nchannels));
    return retcode;
}



int
unstuff_CBBHS (char *srcname, double ipkttime, char *packet, int nbytes, Packet * pkt)
{
    char           *pp,
                   *cp;
    int             retcode = 0;
    int             i;
    double          lcalib;
    float           lsamprate;
    int             lnsamp,
                    ldatatype;
    int             doff;
    short           chlen;
    char            channel_names[ORBSRCNAME_SIZE],
                   *aname;
    char           *lasts;
    int             ICHAN = 0;
    PktChannel     *achan;

    achan = gettbl (pkt->channels, 0);
    if (achan == 0) {
	achan = newPktChannel ();
	pushtbl (pkt->channels, achan);
    }
    pp = cp = packet;
    cp += 4;
    NS2HI (&pkt->version, cp, 1);
    cp += 2 * 1;
    cp += 2;
    switch (pkt->version) {
      case 0xddd:
	cp += PKT_NAMESIZE;
	cp += PKT_NAMESIZE;
	strncpy (channel_names, cp, ORBSRCNAME_SIZE);
	cp += ORBSRCNAME_SIZE;
	N2H4 ((char *) &ldatatype, cp, 1);
	cp += 4 * 1;
	cp += 8;
	N2HD (&lcalib, (uchar *) cp, 1);
	cp += 8 * 1;
	N2H4 ((char *) &lsamprate, cp, 1);
	cp += 4 * 1;
	N2H4 ((char *) &lnsamp, cp, 1);
	cp += 4 * 1;
	N2H4 ((char *) &pkt->nchannels, cp, 1);
	cp += 4 * 1;
	N2H4 ((char *) &doff, cp, 1);
	cp += 4 * 1;
	break;
      default:
	return -1;
	break;
      case 0xbba:
	NF2HD (&lcalib, cp, 1);
	cp += 4 * 1;
	N2H4 ((char *) &lsamprate, cp, 1);
	cp += 4 * 1;
	NS2HI (&ldatatype, cp, 1);
	cp += 2 * 1;
	NS2HI (&lnsamp, cp, 1);
	cp += 2 * 1;
	NS2HI (&pkt->nchannels, cp, 1);
	cp += 2 * 1;
	NS2HI (&doff, cp, 1);
	cp += 2 * 1;
	N2H2 ((char *) &chlen, cp, 1);
	cp += 2 * 1;
	if (chlen <= 0 || chlen > ORBSRCNAME_SIZE)
	    return -1;
	memcpy (&channel_names, cp, chlen);
	cp += 1 * chlen;
	channel_names[chlen] = 0;
	break;
    };
    if (lnsamp < 1 || lnsamp > MAX_NSAMP) {
	complain (0, "lnsamp=%d", lnsamp);
	return -1;
    };
    if (pkt->nchannels > 3 || pkt->nchannels < 1) {
	complain (0, "pkt->nchannels=%d", pkt->nchannels);
	return -1;
    };
    aname = strtok_r (channel_names, "_", &lasts);
    cp += doff;
    for (ICHAN = 0; ICHAN < pkt->nchannels; ICHAN++) {
	achan = gettbl (pkt->channels, ICHAN);
	if (achan == 0) {
	    achan = newPktChannel ();
	    pushtbl (pkt->channels, achan);
	};
	strncpy (achan->net, pkt->parts.src_net, PKT_NAMESIZE);
	strncpy (achan->sta, pkt->parts.src_sta, PKT_NAMESIZE);
	if (aname == 0)
	    return -1;
	strcpy (achan->chan, aname);
	achan->time = ipkttime;
	aname = strtok_r (0, "_", &lasts);
	achan->samprate = lsamprate;
	achan->calib = lcalib;
	strcpy (achan->segtype, "V");
	achan->calper = 0.0;
	achan->nsamp = lnsamp;
	if (achan->nsamp < 0 || achan->nsamp > MAX_NSAMP)
	    return -1;
	RESIZE_BUFFER (int *, achan->data, achan->datasz, achan->nsamp + 1);
	achan->iuser1 = ldatatype;
	switch (achan->iuser1) {
	  case 3:		       /* integer */
	    N2H4 ((char *) achan->data, cp, achan->nsamp);
	    cp += 4 * achan->nsamp;
	    break;
	  case 2:		       /* short */
	    NS2HI (achan->data, cp, achan->nsamp);
	    cp += 2 * achan->nsamp;
	    break;
	  case 0:
	  default:
	    if (achan->iuser1 != 0)
		retcode = -1;
	    else
		retcode += uucsd (achan->data, &cp, nbytes - (cp - pp), achan->nsamp);
	    break;
	}
    };
    if (nbytes < cp - pp) {
	complain (0, "nbytes=%d < cp-pp=%d\n", nbytes, cp - pp);
	retcode = -1;
    };
    return retcode;
}


int
stuff_CBBLS (Packet * pkt, char *srcname, double *opkttime, char **ppp, int *nbytes, int *ppsz)
{
    return stuff_CBBHS (pkt, srcname, opkttime, ppp, nbytes, ppsz);
}

int
unstuff_CBBLS (char *srcname, double ipkttime, char *packet, int nbytes, Packet * pkt)
{
    return unstuff_CBBHS (srcname, ipkttime, packet, nbytes, pkt);
}


int
stuff_CBB1S (Packet * pkt, char *srcname, double *opkttime, char **ppp, int *nbytes, int *ppsz)
{
    return stuff_CBBHS (pkt, srcname, opkttime, ppp, nbytes, ppsz);
}

int
unstuff_CBB1S (char *srcname, double ipkttime, char *packet, int nbytes, Packet * pkt)
{
    return unstuff_CBBHS (srcname, ipkttime, packet, nbytes, pkt);
}


int
stuff_CBB1S_delayed (Packet * pkt, char *srcname, double *opkttime, char **ppp, int *nbytes, int *ppsz)
{
    return stuff_CBBHS (pkt, srcname, opkttime, ppp, nbytes, ppsz);
}

int
unstuff_CBB1S_delayed (char *srcname, double ipkttime, char *packet, int nbytes, Packet * pkt)
{
    return unstuff_CBBHS (srcname, ipkttime, packet, nbytes, pkt);
}


int
stuff_CALS (Packet * pkt, char *srcname, double *opkttime, char **ppp, int *nbytes, int *ppsz)
{
    return stuff_CBBHS (pkt, srcname, opkttime, ppp, nbytes, ppsz);
}

int
unstuff_CALS (char *srcname, double ipkttime, char *packet, int nbytes, Packet * pkt)
{
    return unstuff_CBBHS (srcname, ipkttime, packet, nbytes, pkt);
}


int
stuff_CAHS (Packet * pkt, char *srcname, double *opkttime, char **ppp, int *nbytes, int *ppsz)
{
    return stuff_CBBHS (pkt, srcname, opkttime, ppp, nbytes, ppsz);
}

int
unstuff_CAHS (char *srcname, double ipkttime, char *packet, int nbytes, Packet * pkt)
{
    return unstuff_CBBHS (srcname, ipkttime, packet, nbytes, pkt);
}


int
stuff_B3S2DP (Packet * pkt, char *srcname, double *opkttime, char **ppp, int *nbytes, int *ppsz)
{
    char           *pp,
                   *cp;
    int             retcode = 0;
    int             doff,
                    ldatatype,
                    lnsamp;
    double          ltime,
                    lcalib;
    float           lsamprate;
    int             hdrsize,
                    pktsize,
                    hdrtype,
                    pkttype;
    int             ICHAN = 0;
    PktChannel     *achan;
    achan = gettbl (pkt->channels, 0);
    pp = cp = *ppp;
    RESIZE_BUFFER (char *, *ppp, *ppsz, (200));
    cp = *ppp + (cp - pp);
    pp = *ppp;
    H2N2 (cp, (char *) &Zero, 1);
    cp += 2 * 1;
    H2N2 (cp, (char *) &Zero, 1);
    cp += 2 * 1;
    HI2NS (cp, (int *) &pkt->pkttype->hdrcode, 1);
    cp += 2 * 1;
    HI2NS (cp, (int *) &pkt->pkttype->bodycode, 1);
    cp += 2 * 1;
    *opkttime = achan->time;
    retcode = -1;
    *nbytes = cp - pp;
    insist (*nbytes < (200));
    return retcode;
}



int
unstuff_B3S2DP (char *srcname, double ipkttime, char *packet, int nbytes, Packet * pkt)
{
    char           *pp,
                   *cp;
    int             retcode = 0;
    int             doff,
                    ldatatype,
                    lnsamp;
    double          ltime,
                    lcalib;
    float           lsamprate;
    int             hdrsize,
                    pktsize,
                    hdrtype,
                    pkttype;
    int             ICHAN = 0;
    PktChannel     *achan;

    achan = gettbl (pkt->channels, 0);
    if (achan == 0) {
	achan = newPktChannel ();
	pushtbl (pkt->channels, achan);
    }
    pp = cp = packet;
    strncpy (achan->net, pkt->parts.src_net, PKT_NAMESIZE);
    strncpy (achan->sta, pkt->parts.src_sta, PKT_NAMESIZE);
    strncpy (achan->chan, pkt->parts.src_chan, PKT_NAMESIZE);
    NS2HI (&hdrsize, cp, 1);
    cp += 2 * 1;
    NS2HI (&pktsize, cp, 1);
    cp += 2 * 1;
    NS2HI (&hdrtype, cp, 1);
    cp += 2 * 1;
    NS2HI (&pkttype, cp, 1);
    cp += 2 * 1;
    cp += 8;
    achan->nsamp = 20;
    achan->time = ipkttime;
    achan->samprate = 20;
    pkt->nchannels = 1;
    RESIZE_BUFFER (int *, achan->data, achan->datasz, achan->nsamp + 1);
    N32H4 (achan->data, pp + nbytes - 60, achan->nsamp);
    if (nbytes < cp - pp) {
	complain (0, "nbytes=%d < cp-pp=%d\n", nbytes, cp - pp);
	retcode = -1;
    };
    return retcode;
}


int
stuff_BSP (Packet * pkt, char *srcname, double *opkttime, char **ppp, int *nbytes, int *ppsz)
{
    char           *pp,
                   *cp;
    int             retcode = 0;
    pp = cp = *ppp;
    RESIZE_BUFFER (char *, *ppp, *ppsz, 24);
    cp = *ppp + (cp - pp);
    pp = *ppp;
    *nbytes = cp - pp;
    insist (*nbytes < 24);
    return retcode;
}



int
unstuff_BSP (char *srcname, double ipkttime, char *packet, int nbytes, Packet * pkt)
{
    char           *pp,
                   *cp;
    int             retcode = 0;
    pp = cp = packet;
    if (nbytes < cp - pp) {
	complain (0, "nbytes=%d < cp-pp=%d\n", nbytes, cp - pp);
	retcode = -1;
    };
    return retcode;
}


int
stuff_LISS (Packet * pkt, char *srcname, double *opkttime, char **ppp, int *nbytes, int *ppsz)
{
    char           *pp,
                   *cp;
    int             retcode = 0;
    int             new_version = 2;
    static char    *V = "V";
    int             ICHAN = 0;
    PktChannel     *achan;
    achan = gettbl (pkt->channels, 0);
    SIZE_BUFFER (char *, *ppp, *ppsz, (8 + (15 * pkt->nchannels) + 5 * pkt->nchannels * achan->nsamp));
    pp = cp = *ppp;
#define LISS_DATA_HDR_TYPE 0x6c6c
    H2N2 (cp, (char *) &Zero, 1);
    cp += 2 * 1;
    H2N2 (cp, (char *) &Zero, 1);
    cp += 2 * 1;
    HI2NS (cp, (int *) &pkt->pkttype->hdrcode, 1);
    cp += 2 * 1;
    HI2NS (cp, (int *) &new_version, 1);
    cp += 2 * 1;
    HI2NS (cp, (int *) &achan->nsamp, 1);
    cp += 2 * 1;
    cp += 2;
    HD2NF (cp, &achan->samprate, 1);
    cp += 4 * 1;
    *opkttime = achan->time;
    switch (pkt->version) {
      case LISS_DATA_HDR_TYPE:
	break;
      case 2:
	HD2NF (cp, &achan->calib, 1);
	cp += 4 * 1;
	HD2NF (cp, &achan->calper, 1);
	cp += 4 * 1;
	memcpy (cp, &achan->segtype, 1);
	cp += 1 * 1;
	break;
      default:
	retcode = -1;
	break;
    }
    retcode = -1;
    *nbytes = cp - pp;
    insist (*nbytes < (8 + (15 * pkt->nchannels) + 5 * pkt->nchannels * achan->nsamp));
    return retcode;
}



int
unstuff_LISS (char *srcname, double ipkttime, char *packet, int nbytes, Packet * pkt)
{
    char           *pp,
                   *cp;
    int             retcode = 0;
    static char    *V = "V";
    int             ICHAN = 0;
    PktChannel     *achan;

    achan = gettbl (pkt->channels, 0);
    if (achan == 0) {
	achan = newPktChannel ();
	pushtbl (pkt->channels, achan);
    }
    pp = cp = packet;
#define LISS_DATA_HDR_TYPE 0x6c6c
    cp += 6;
    NS2HI (&pkt->version, cp, 1);
    cp += 2 * 1;
    NS2HI (&achan->nsamp, cp, 1);
    cp += 2 * 1;
    if (achan->nsamp < 0 || achan->nsamp > MAX_SEED_NSAMP)
	return -1;
    cp += 2;
    NF2HD (&achan->samprate, cp, 1);
    cp += 4 * 1;
    pkt->nchannels = 1;
    RESIZE_BUFFER (int *, achan->data, achan->datasz, achan->nsamp + 1);
    strncpy (achan->net, pkt->parts.src_net, PKT_NAMESIZE);
    strncpy (achan->sta, pkt->parts.src_sta, PKT_NAMESIZE);
    strncpy (achan->chan, pkt->parts.src_chan, PKT_NAMESIZE);
    strncpy (achan->loc, pkt->parts.src_loc, PKT_NAMESIZE);
    switch (pkt->version) {
      case LISS_DATA_HDR_TYPE:
	strncpy (achan->segtype, V, 4);
	break;
      case 2:
	NF2HD (&achan->calib, cp, 1);
	cp += 4 * 1;
	NF2HD (&achan->calper, cp, 1);
	cp += 4 * 1;
	memcpy (&achan->segtype, cp, 1);
	cp += 1 * 1;
	break;
      default:
	retcode = -1;
	break;
    }
    retcode = deseed (cp, nbytes - (cp - pp), &achan->time, &achan->samprate, &achan->nsamp, &achan->data, &achan->datasz, &pkt->pkthook);
    achan->time = ipkttime;
    if (nbytes < cp - pp) {
	complain (0, "nbytes=%d < cp-pp=%d\n", nbytes, cp - pp);
	retcode = -1;
    };
    return retcode;
}


int
stuff_SEED (Packet * pkt, char *srcname, double *opkttime, char **ppp, int *nbytes, int *ppsz)
{
    char           *pp,
                   *cp;
    int             retcode = 0;
    int             ICHAN = 0;
    PktChannel     *achan;
    achan = gettbl (pkt->channels, 0);
    pp = cp = *ppp;
    RESIZE_BUFFER (char *, *ppp, *ppsz, 8192);
    cp = *ppp + (cp - pp);
    pp = *ppp;
    HI2NC (cp, &pkt->version, 1);
    cp += 1 * 1;
    memcpy (cp, &achan->segtype, 1);
    cp += 1 * 1;
    HD2NF (cp, &achan->samprate, 1);
    cp += 4 * 1;
    HD2NF (cp, &achan->calib, 1);
    cp += 4 * 1;
    HD2NF (cp, &achan->calper, 1);
    cp += 4 * 1;
    strncpy (pkt->parts.src_net, achan->net, PKT_NAMESIZE);
    strncpy (pkt->parts.src_sta, achan->sta, PKT_NAMESIZE);
    strncpy (pkt->parts.src_chan, achan->chan, PKT_NAMESIZE);
    strncpy (pkt->parts.src_loc, achan->loc, PKT_NAMESIZE);
    *opkttime = achan->time;
    retcode = -1;
    *nbytes = cp - pp;
    insist (*nbytes < 8192);
    return retcode;
}



int
unstuff_SEED (char *srcname, double ipkttime, char *packet, int nbytes, Packet * pkt)
{
    char           *pp,
                   *cp;
    int             retcode = 0;
    int             ICHAN = 0;
    PktChannel     *achan;

    achan = gettbl (pkt->channels, 0);
    if (achan == 0) {
	achan = newPktChannel ();
	pushtbl (pkt->channels, achan);
    }
    pp = cp = packet;
    NC2HI (&pkt->version, cp, 1);
    cp += 1 * 1;
    memcpy (&achan->segtype, cp, 1);
    cp += 1 * 1;
    NF2HD (&achan->samprate, cp, 1);
    cp += 4 * 1;
    NF2HD (&achan->calib, cp, 1);
    cp += 4 * 1;
    NF2HD (&achan->calper, cp, 1);
    cp += 4 * 1;
    pkt->nchannels = 1;
    retcode = deseed (cp, nbytes - (cp - pp), &achan->time, &achan->samprate, &achan->nsamp, &achan->data, &achan->datasz, &pkt->pkthook);
    achan->time = ipkttime;
    strncpy (achan->net, pkt->parts.src_net, PKT_NAMESIZE);
    strncpy (achan->sta, pkt->parts.src_sta, PKT_NAMESIZE);
    strncpy (achan->chan, pkt->parts.src_chan, PKT_NAMESIZE);
    strncpy (achan->loc, pkt->parts.src_loc, PKT_NAMESIZE);
    if (nbytes < cp - pp) {
	complain (0, "nbytes=%d < cp-pp=%d\n", nbytes, cp - pp);
	retcode = -1;
    };
    return retcode;
}


int
stuff_MSEED (Packet * pkt, char *srcname, double *opkttime, char **ppp, int *nbytes, int *ppsz)
{
    char           *pp,
                   *cp;
    int             retcode = 0;
    int             ICHAN = 0;
    PktChannel     *achan;
    achan = gettbl (pkt->channels, 0);
    pp = cp = *ppp;
    RESIZE_BUFFER (char *, *ppp, *ppsz, 8192);
    cp = *ppp + (cp - pp);
    pp = *ppp;
    cp += 8;
    strncpy (pkt->parts.src_net, achan->net, PKT_NAMESIZE);
    strncpy (pkt->parts.src_sta, achan->sta, PKT_NAMESIZE);
    strncpy (pkt->parts.src_chan, achan->chan, PKT_NAMESIZE);
    strncpy (pkt->parts.src_loc, achan->loc, PKT_NAMESIZE);
    *opkttime = achan->time;
    retcode = -1;
    *nbytes = cp - pp;
    insist (*nbytes < 8192);
    return retcode;
}



int
unstuff_MSEED (char *srcname, double ipkttime, char *packet, int nbytes, Packet * pkt)
{
    char           *pp,
                   *cp;
    int             retcode = 0;
    int             ICHAN = 0;
    PktChannel     *achan;

    achan = gettbl (pkt->channels, 0);
    if (achan == 0) {
	achan = newPktChannel ();
	pushtbl (pkt->channels, achan);
    }
    pp = cp = packet;
    cp += 8;
    pkt->nchannels = 1;
    retcode = deseed (cp, nbytes - (cp - pp), &achan->time, &achan->samprate, &achan->nsamp, &achan->data, &achan->datasz, &pkt->pkthook);
    if (retcode == Pkt_ch) {
	if (pkt->string)
	    free (pkt->string);
	if (achan->nsamp > 0)
	    pkt->string_size = achan->nsamp;
	else
	    pkt->string_size = MIN (nbytes - 64, MAX_CH_SIZE);
	allot (char *, pkt->string, pkt->string_size + 1);
	memcpy (pkt->string, pp + 64, pkt->string_size);
	pkt->string[pkt->string_size] = 0;
	cp = pp + 64 + pkt->string_size;
    };
    achan->time = ipkttime;
    strncpy (achan->net, pkt->parts.src_net, PKT_NAMESIZE);
    strncpy (achan->sta, pkt->parts.src_sta, PKT_NAMESIZE);
    strncpy (achan->chan, pkt->parts.src_chan, PKT_NAMESIZE);
    strncpy (achan->loc, pkt->parts.src_loc, PKT_NAMESIZE);
    if (nbytes < cp - pp) {
	complain (0, "nbytes=%d < cp-pp=%d\n", nbytes, cp - pp);
	retcode = -1;
    };
    return retcode;
}


int
stuff_raw (Packet * pkt, char *srcname, double *opkttime, char **ppp, int *nbytes, int *ppsz)
{
    char           *pp,
                   *cp;
    int             retcode = 0;
    pp = cp = *ppp;
    RESIZE_BUFFER (char *, *ppp, *ppsz, 10);
    cp = *ppp + (cp - pp);
    pp = *ppp;
    *opkttime = now ();
    *nbytes = cp - pp;
    insist (*nbytes < 10);
    return retcode;
}



int
unstuff_raw (char *srcname, double ipkttime, char *packet, int nbytes, Packet * pkt)
{
    char           *pp,
                   *cp;
    int             retcode = 0;
    pp = cp = packet;
    if (nbytes < cp - pp) {
	complain (0, "nbytes=%d < cp-pp=%d\n", nbytes, cp - pp);
	retcode = -1;
    };
    return retcode;
}


int
stuff_file (Packet * pkt, char *srcname, double *opkttime, char **ppp, int *nbytes, int *ppsz)
{
    return stuff_raw (pkt, srcname, opkttime, ppp, nbytes, ppsz);
}

int
unstuff_file (char *srcname, double ipkttime, char *packet, int nbytes, Packet * pkt)
{
    return unstuff_raw (srcname, ipkttime, packet, nbytes, pkt);
}


int
stuff_xfer (Packet * pkt, char *srcname, double *opkttime, char **ppp, int *nbytes, int *ppsz)
{
    char           *pp,
                   *cp;
    int             retcode = 0;
    pp = cp = *ppp;
#define DIGEST_SIZE 16
    RESIZE_BUFFER (char *, *ppp, *ppsz, pkt->dfile_size + pkt->string_size + DIGEST_SIZE + 20);
    cp = *ppp + (cp - pp);
    pp = *ppp;
    HI2NC (cp, &pkt->version, 1);
    cp += 1 * 1;
    H2N4 (cp, (char *) &pkt->nchannels, 1);
    cp += 4 * 1;
    H2N4 (cp, (char *) &pkt->string_size, 1);
    cp += 4 * 1;
    memcpy (cp, pkt->string, DIGEST_SIZE);
    cp += DIGEST_SIZE;
    H2N4 (cp, (char *) &pkt->dfile_size, 1);
    cp += 4 * 1;
    if (pkt->dfile_size != 0) {
	memcpy (cp, pkt->dfile, pkt->dfile_size);
	cp += pkt->dfile_size;
    }
    *nbytes = cp - pp;
    insist (*nbytes < pkt->dfile_size + pkt->string_size + DIGEST_SIZE + 20);
    return retcode;
}



int
unstuff_xfer (char *srcname, double ipkttime, char *packet, int nbytes, Packet * pkt)
{
    char           *pp,
                   *cp;
    int             retcode = 0;
    pp = cp = packet;
#define DIGEST_SIZE 16
    NC2HI (&pkt->version, cp, 1);
    cp += 1 * 1;
    N2H4 ((char *) &pkt->nchannels, cp, 1);
    cp += 4 * 1;
    N2H4 ((char *) &pkt->string_size, cp, 1);
    cp += 4 * 1;
    if (pkt->string)
	free (pkt->string);
    allot (char *, pkt->string, DIGEST_SIZE + 4);
    memcpy (pkt->string, cp, DIGEST_SIZE);
    pkt->string[DIGEST_SIZE] = 0;
    cp += DIGEST_SIZE;
    N2H4 ((char *) &pkt->dfile_size, cp, 1);
    cp += 4 * 1;
    if (pkt->dfile_size != 0) {
	if (pkt->dfile)
	    free (pkt->dfile);
	allot (char *, pkt->dfile, pkt->dfile_size);
	memcpy (pkt->dfile, cp, pkt->dfile_size);
    }
    if (nbytes < cp - pp) {
	complain (0, "nbytes=%d < cp-pp=%d\n", nbytes, cp - pp);
	retcode = -1;
    };
    return retcode;
}


int
stuff_QCDAT (Packet * pkt, char *srcname, double *opkttime, char **ppp, int *nbytes, int *ppsz)
{
    char           *pp,
                   *cp;
    int             retcode = 0;
    static char    *V = "V";
    int             ICHAN = 0;
    PktChannel     *achan;
    achan = gettbl (pkt->channels, 0);
    SIZE_BUFFER (char *, *ppp, *ppsz, (8 + (28 * pkt->nchannels) + 5 * pkt->nchannels * achan->nsamp));
    pp = cp = *ppp;
#define QUANTERRA_DATA_HDR_TYPE2 0x7165
    H2N2 (cp, (char *) &Zero, 1);
    cp += 2 * 1;
    H2N2 (cp, (char *) &Zero, 1);
    cp += 2 * 1;
    HI2NS (cp, (int *) &pkt->pkttype->hdrcode, 1);
    cp += 2 * 1;
    HI2NS (cp, (int *) &pkt->pkttype->bodycode, 1);
    cp += 2 * 1;
    memcpy (cp, achan->cuser1, 2);
    cp += 1 * 2;
    HI2NS (cp, (int *) &achan->iuser1, 1);
    cp += 2 * 1;
    H2N4 (cp, (char *) &achan->nsamp, 1);
    cp += 4 * 1;
    HD2NF (cp, &achan->samprate, 1);
    cp += 4 * 1;
    HD2NF (cp, &achan->calib, 1);
    cp += 4 * 1;
    HD2NF (cp, &achan->calper, 1);
    cp += 4 * 1;
    memcpy (cp, achan->cuser2, 4);
    cp += 1 * 4;
    switch (pkt->version) {
      case QUANTERRA_DATA_HDR_TYPE2:
	memcpy (cp, &achan->segtype, 4);
	cp += 1 * 4;
	break;
      default:
	break;
    }
    RESIZE_BUFFER (int *, achan->data, achan->datasz, achan->nsamp + 1);
    *opkttime = achan->time;
    retcode = -1;
    *nbytes = cp - pp;
    insist (*nbytes < (8 + (28 * pkt->nchannels) + 5 * pkt->nchannels * achan->nsamp));
    return retcode;
}



int
unstuff_QCDAT (char *srcname, double ipkttime, char *packet, int nbytes, Packet * pkt)
{
    char           *pp,
                   *cp;
    int             retcode = 0;
    static char    *V = "V";
    int             ICHAN = 0;
    PktChannel     *achan;

    achan = gettbl (pkt->channels, 0);
    if (achan == 0) {
	achan = newPktChannel ();
	pushtbl (pkt->channels, achan);
    }
    pp = cp = packet;
#define QUANTERRA_DATA_HDR_TYPE2 0x7165
    cp += 6;
    NS2HI (&pkt->version, cp, 1);
    cp += 2 * 1;
    memcpy (achan->cuser1, cp, 2);
    cp += 1 * 2;
    NS2HI (&achan->iuser1, cp, 1);
    cp += 2 * 1;
    N2H4 ((char *) &achan->nsamp, cp, 1);
    cp += 4 * 1;
    if (achan->nsamp < 0 || achan->nsamp > MAX_NSAMP)
	return -1;
    NF2HD (&achan->samprate, cp, 1);
    cp += 4 * 1;
    NF2HD (&achan->calib, cp, 1);
    cp += 4 * 1;
    NF2HD (&achan->calper, cp, 1);
    cp += 4 * 1;
    memcpy (achan->cuser2, cp, 4);
    cp += 1 * 4;
    switch (pkt->version) {
      case QUANTERRA_DATA_HDR_TYPE2:
	memcpy (&achan->segtype, cp, 4);
	cp += 1 * 4;
	break;
      default:
	strncpy (achan->segtype, V, 4);
	break;
    }
    pkt->nchannels = 1;
    RESIZE_BUFFER (int *, achan->data, achan->datasz, achan->nsamp + 1);
    achan->time = ipkttime;
    strncpy (achan->net, pkt->parts.src_net, PKT_NAMESIZE);
    strncpy (achan->sta, pkt->parts.src_sta, PKT_NAMESIZE);
    strncpy (achan->chan, pkt->parts.src_chan, PKT_NAMESIZE);
    strncpy (achan->loc, pkt->parts.src_loc, PKT_NAMESIZE);
    retcode = unsteim (achan->data, achan->nsamp, cp, nbytes - (cp - pp), achan->iuser1);
    if (nbytes < cp - pp) {
	complain (0, "nbytes=%d < cp-pp=%d\n", nbytes, cp - pp);
	retcode = -1;
    };
    return retcode;
}


int
stuff_CD1S (Packet * pkt, char *srcname, double *opkttime, char **ppp, int *nbytes, int *ppsz)
{
    char           *pp,
                   *cp;
    int             retcode = 0;
    int             cd1_length,
                    cd1_nsamp;
    double          cd1_time;
    char            cd1_status[4];
    int             nauth,
                    nb,
                    datsz;
    char           *tmp;
    int             maxsz;
    int             ICHAN = 0;
    PktChannel     *achan;
    achan = gettbl (pkt->channels, 0);
    pp = cp = *ppp;
    RESIZE_BUFFER (char *, *ppp, *ppsz, 20 + 60 + (achan->nsamp + 1) * 10);
    cp = *ppp + (cp - pp);
    pp = *ppp;
    maxsz = 20 + 60 + (achan->nsamp + 1) * 10;
    memset (pp, 0, maxsz);
    HI2NC (cp, &pkt->version, 1);
    cp += 1 * 1;
    memcpy (cp, &achan->segtype, 1);
    cp += 1 * 1;
    HI2NC (cp, &achan->iuser1, 1);
    cp += 1 * 1;
    HI2NC (cp, &achan->iuser2, 1);
    cp += 1 * 1;
    HD2NF (cp, &achan->samprate, 1);
    cp += 4 * 1;
    HD2NF (cp, &achan->calib, 1);
    cp += 4 * 1;
    HD2NF (cp, &achan->calper, 1);
    cp += 4 * 1;
    HD2NF (cp, &achan->duser1, 1);
    cp += 4 * 1;
    cd1_time = achan->time;
    cd1_nsamp = achan->nsamp;
    cd1_status[0] = achan->cuser1[0];
    cd1_status[1] = achan->cuser1[1];
    cd1_status[2] = achan->cuser1[2];
    cd1_status[3] = achan->cuser1[3];
    if (achan->iuser2) {
	cd1_length = 56;
    } else {
	cd1_length = 16;
    };
    tmp = cp;
    H2N4 (cp, (char *) &cd1_length, 1);
    cp += 4 * 1;
    if (achan->iuser2) {
	memcpy (cp, achan->cuser2, 40);
	cp += 1 * 40;
    };
    H2ND ((uchar *) cp, &cd1_time, 1);
    cp += 8 * 1;
    H2N4 (cp, (char *) &cd1_nsamp, 1);
    cp += 4 * 1;
    memcpy (cp, &cd1_status, 4);
    cp += 1 * 4;
    retcode = 0;
    if (achan->nsamp > 0) {
	if (achan->iuser1) {
	    nb = maxsz - (cp - pp);
	    retcode = cmcanada (achan->data, achan->nsamp, cp, &nb);
	    cp += nb;
	} else {
	    H2N4 (cp, (char *) achan->data, achan->nsamp);
	    cp += 4 * achan->nsamp;
	};
    };
    cd1_length = cp - tmp - 4;
    if (cd1_length % 4) {
	nauth = ((cd1_length - 1) / 4 + 1) * 4;
	cp += nauth - cd1_length;
	cd1_length = nauth;
    };
    H2N4 (tmp, (char *) &cd1_length, 1);
    *nbytes = cp - pp;
    insist (*nbytes < 20 + 60 + (achan->nsamp + 1) * 10);
    return retcode;
}



int
unstuff_CD1S (char *srcname, double ipkttime, char *packet, int nbytes, Packet * pkt)
{
    char           *pp,
                   *cp;
    int             retcode = 0;
    int             cd1_length,
                    cd1_nsamp;
    double          cd1_time;
    char            cd1_status[4];
    int             nauth,
                    nb,
                    datsz;
    char           *tmp;
    int             maxsz;
    int             ICHAN = 0;
    PktChannel     *achan;

    achan = gettbl (pkt->channels, 0);
    if (achan == 0) {
	achan = newPktChannel ();
	pushtbl (pkt->channels, achan);
    }
    pp = cp = packet;
    NC2HI (&pkt->version, cp, 1);
    cp += 1 * 1;
    memcpy (&achan->segtype, cp, 1);
    cp += 1 * 1;
    NC2HI (&achan->iuser1, cp, 1);
    cp += 1 * 1;
    NC2HI (&achan->iuser2, cp, 1);
    cp += 1 * 1;
    NF2HD (&achan->samprate, cp, 1);
    cp += 4 * 1;
    NF2HD (&achan->calib, cp, 1);
    cp += 4 * 1;
    NF2HD (&achan->calper, cp, 1);
    cp += 4 * 1;
    NF2HD (&achan->duser1, cp, 1);
    cp += 4 * 1;
    pkt->nchannels = 1;
    N2H4 ((char *) &cd1_length, cp, 1);
    cp += 4 * 1;
    if (achan->iuser2) {
	memcpy (achan->cuser2, cp, 40);
	cp += 1 * 40;
	nauth = 40;
    } else {
	nauth = 0;
    };
    N2HD (&cd1_time, (uchar *) cp, 1);
    cp += 8 * 1;
    achan->time = cd1_time;
    N2H4 ((char *) &cd1_nsamp, cp, 1);
    cp += 4 * 1;
    if (cd1_nsamp < 0 || cd1_nsamp > MAX_CD1_NSAMP)
	return -1;
    achan->nsamp = cd1_nsamp + 2;
    RESIZE_BUFFER (int *, achan->data, achan->datasz, achan->nsamp + 1);
    datsz = achan->nsamp * sizeof (int);
    achan->nsamp -= 2;
    memcpy (&cd1_status, cp, 4);
    cp += 1 * 4;
    retcode = 0;
    if (achan->nsamp > 0) {
	achan->puser1 = cp;
	if (achan->iuser1) {
	    nb = cd1_length - 16 - nauth;
	    achan->iuser3 = nb;
	    retcode = uncanada (achan->data, achan->nsamp, datsz, cp, nb);
	    cp += nb;
	} else {
	    nb = cd1_length - 16 - nauth;
	    if (nb == 3 * achan->nsamp) {
		N32H4 (achan->data, cp, achan->nsamp);
	    } else if (nb == 4 * achan->nsamp) {
		N2H4 ((char *) achan->data, cp, achan->nsamp);
	    } else if (nb == 2 * achan->nsamp) {
		NS2HI (achan->data, cp, achan->nsamp);
	    } else {
		return -1;
	    };
	    cp += nb;
	    achan->iuser3 = nb;
	};
    };
    strncpy (achan->net, pkt->parts.src_net, PKT_NAMESIZE);
    strncpy (achan->sta, pkt->parts.src_sta, PKT_NAMESIZE);
    strncpy (achan->chan, pkt->parts.src_chan, PKT_NAMESIZE);
    strncpy (achan->loc, pkt->parts.src_loc, PKT_NAMESIZE);
    achan->cuser1[0] = cd1_status[0];
    achan->cuser1[1] = cd1_status[1];
    achan->cuser1[2] = cd1_status[2];
    achan->cuser1[3] = cd1_status[3];
    if (nbytes < cp - pp) {
	complain (0, "nbytes=%d < cp-pp=%d\n", nbytes, cp - pp);
	retcode = -1;
    };
    return retcode;
}


int
stuff_CD1 (Packet * pkt, char *srcname, double *opkttime, char **ppp, int *nbytes, int *ppsz)
{
    char           *pp,
                   *cp;
    int             retcode = 0;
    int             cd1_length,
                    cd1_nsamp;
    double          cd1_time,
                    cd1_twin;
    char            cd1_status[4];
    int             nauth,
                    nb,
                    datsz;
    char           *ftmp,
                   *tmp;
    int             maxsz;
    int             first = 1;
    int             cd1_frameid = 1;
    int             cd1_framelength = 0;
    int             cd1_descrl = 0;
    double          cd1_nomtime = 0.0;
    char            cd1_descr[1024];
    int             ICHAN = 0;
    PktChannel     *achan;
    achan = gettbl (pkt->channels, 0);
    pp = cp = *ppp;
    maxsz = 30;
    for (ICHAN = 0; ICHAN < pkt->nchannels; ICHAN++) {
	achan = gettbl (pkt->channels, ICHAN);
	if (achan == 0) {
	    achan = newPktChannel ();
	    pushtbl (pkt->channels, achan);
	};
	maxsz += 28 + 60 + (achan->nsamp + 1) * 10;
	if (first) {
	    cd1_twin = achan->duser1;
	    first = 0;
	} else if (achan->duser1 != cd1_twin) {
	    register_error (0, "stuff_CD1: twin must be the same across pktchans.");
	    return (-1);
	};
    };
    if (pkt->string != 0)
	cd1_descrl = ((int) ((strlen (pkt->string) + 3) / 4) * 4);
    else
	cd1_descrl = 0;
    maxsz += cd1_descrl;
    RESIZE_BUFFER (char *, *ppp, *ppsz, maxsz);
    cp = *ppp + (cp - pp);
    pp = *ppp;
    memset (pp, 0, maxsz);
    HI2NC (cp, &pkt->version, 1);
    cp += 1 * 1;
    HD2NF (cp, &cd1_twin, 1);
    cp += 4 * 1;
    H2N4 (cp, (char *) &pkt->nchannels, 1);
    cp += 4 * 1;
    for (ICHAN = 0; ICHAN < pkt->nchannels; ICHAN++) {
	achan = gettbl (pkt->channels, ICHAN);
	if (achan == 0) {
	    achan = newPktChannel ();
	    pushtbl (pkt->channels, achan);
	};
	HD2NF (cp, &achan->samprate, 1);
	cp += 4 * 1;
	HD2NF (cp, &achan->calib, 1);
	cp += 4 * 1;
	HD2NF (cp, &achan->calper, 1);
	cp += 4 * 1;
	zstrcopy ((char *) cp, (char *) &achan->sta, 6);
	cp += 1 * 6;
	zstrcopy ((char *) cp, (char *) &achan->chan, 6);
	cp += 1 * 6;
	memcpy (cp, &achan->segtype, 1);
	cp += 1 * 1;
	HI2NC (cp, &achan->iuser1, 1);
	cp += 1 * 1;
	HI2NC (cp, &achan->iuser2, 1);
	cp += 1 * 1;
	cd1_nomtime = MAX (floor (achan->time + 0.5 / achan->samprate), cd1_nomtime);
    };
    H2N4 (cp, (char *) &cd1_frameid, 1);
    cp += 4 * 1;
    ftmp = cp;
    H2N4 (cp, (char *) &cd1_framelength, 1);
    cp += 4 * 1;
    *opkttime = cd1_nomtime;
    H2ND ((uchar *) cp, &cd1_nomtime, 1);
    cp += 8 * 1;
    H2N4 (cp, (char *) &cd1_descrl, 1);
    cp += 4 * 1;
    if (cd1_descrl > 0) {
	zstrcopy ((char *) cp, (char *) pkt->string, cd1_descrl);
	cp += 1 * cd1_descrl;
    };
    for (ICHAN = 0; ICHAN < pkt->nchannels; ICHAN++) {
	achan = gettbl (pkt->channels, ICHAN);
	if (achan == 0) {
	    achan = newPktChannel ();
	    pushtbl (pkt->channels, achan);
	};
	strncpy (pkt->parts.src_net, achan->net, PKT_NAMESIZE);
	cd1_time = achan->time;
	cd1_nsamp = achan->nsamp;
	cd1_status[0] = achan->cuser1[0];
	cd1_status[1] = achan->cuser1[1];
	cd1_status[2] = achan->cuser1[2];
	cd1_status[3] = achan->cuser1[3];
	if (achan->iuser2) {
	    cd1_length = 56;
	} else {
	    cd1_length = 16;
	};
	tmp = cp;
	H2N4 (cp, (char *) &cd1_length, 1);
	cp += 4 * 1;
	if (achan->iuser2) {
	    memcpy (cp, achan->cuser2, 40);
	    cp += 1 * 40;
	};
	H2ND ((uchar *) cp, &cd1_time, 1);
	cp += 8 * 1;
	H2N4 (cp, (char *) &cd1_nsamp, 1);
	cp += 4 * 1;
	memcpy (cp, &cd1_status, 4);
	cp += 1 * 4;
	retcode = 0;
	if (achan->nsamp > 0) {
	    if (achan->iuser1) {
		nb = maxsz - (cp - pp);
		retcode = cmcanada (achan->data, achan->nsamp, cp, &nb);
		cp += nb;
	    } else {
		H2N4 (cp, (char *) achan->data, achan->nsamp);
		cp += 4 * achan->nsamp;
	    };
	};
	cd1_length = cp - tmp - 4;
	if (cd1_length % 4) {
	    nauth = ((cd1_length - 1) / 4 + 1) * 4;
	    cp += nauth - cd1_length;
	    cd1_length = nauth;
	};
	H2N4 (tmp, (char *) &cd1_length, 1);
    };
    cd1_framelength = cp - ftmp + 4;
    H2N4 (ftmp, (char *) &cd1_framelength, 1);
    *nbytes = cp - pp;
    insist (*nbytes < maxsz);
    return retcode;
}



int
unstuff_CD1 (char *srcname, double ipkttime, char *packet, int nbytes, Packet * pkt)
{
    char           *pp,
                   *cp;
    int             retcode = 0;
    int             cd1_length,
                    cd1_nsamp;
    double          cd1_time,
                    cd1_twin;
    char            cd1_status[4];
    int             nauth,
                    nb,
                    datsz;
    char           *ftmp,
                   *tmp;
    int             maxsz;
    int             first = 1;
    int             cd1_frameid = 1;
    int             cd1_framelength = 0;
    int             cd1_descrl = 0;
    double          cd1_nomtime = 0.0;
    char            cd1_descr[1024];
    int             ICHAN = 0;
    PktChannel     *achan;

    achan = gettbl (pkt->channels, 0);
    if (achan == 0) {
	achan = newPktChannel ();
	pushtbl (pkt->channels, achan);
    }
    pp = cp = packet;
    NC2HI (&pkt->version, cp, 1);
    cp += 1 * 1;
    NF2HD (&cd1_twin, cp, 1);
    cp += 4 * 1;
    N2H4 ((char *) &pkt->nchannels, cp, 1);
    cp += 4 * 1;
    if (pkt->nchannels < 1 || pkt->nchannels > 100)
	return -1;
    for (ICHAN = 0; ICHAN < pkt->nchannels; ICHAN++) {
	achan = gettbl (pkt->channels, ICHAN);
	if (achan == 0) {
	    achan = newPktChannel ();
	    pushtbl (pkt->channels, achan);
	};
	NF2HD (&achan->samprate, cp, 1);
	cp += 4 * 1;
	NF2HD (&achan->calib, cp, 1);
	cp += 4 * 1;
	NF2HD (&achan->calper, cp, 1);
	cp += 4 * 1;
	zstrcopy ((char *) &achan->sta, (char *) cp, 6);
	cp += 1 * 6;
	zstrcopy ((char *) &achan->chan, (char *) cp, 6);
	cp += 1 * 6;
	memcpy (&achan->segtype, cp, 1);
	cp += 1 * 1;
	NC2HI (&achan->iuser1, cp, 1);
	cp += 1 * 1;
	NC2HI (&achan->iuser2, cp, 1);
	cp += 1 * 1;
	achan->duser1 = cd1_twin;
    };
    N2H4 ((char *) &cd1_frameid, cp, 1);
    cp += 4 * 1;
    N2H4 ((char *) &cd1_framelength, cp, 1);
    cp += 4 * 1;
    N2HD (&cd1_nomtime, (uchar *) cp, 1);
    cp += 8 * 1;
    N2H4 ((char *) &cd1_descrl, cp, 1);
    cp += 4 * 1;
    if (cd1_descrl > 0) {
	if (pkt->string != 0)
	    free (pkt->string);
	allot (char *, pkt->string, cd1_descrl);
	zstrcopy ((char *) pkt->string, (char *) cp, cd1_descrl);
	cp += 1 * cd1_descrl;
    };
    for (ICHAN = 0; ICHAN < pkt->nchannels; ICHAN++) {
	achan = gettbl (pkt->channels, ICHAN);
	if (achan == 0) {
	    achan = newPktChannel ();
	    pushtbl (pkt->channels, achan);
	};
	N2H4 ((char *) &cd1_length, cp, 1);
	cp += 4 * 1;
	if (achan->iuser2) {
	    memcpy (achan->cuser2, cp, 40);
	    cp += 1 * 40;
	    nauth = 40;
	} else {
	    nauth = 0;
	};
	N2HD (&cd1_time, (uchar *) cp, 1);
	cp += 8 * 1;
	achan->time = cd1_time;
	N2H4 ((char *) &cd1_nsamp, cp, 1);
	cp += 4 * 1;
	if (cd1_nsamp < 0 || cd1_nsamp > MAX_CD1_NSAMP)
	    return -1;
	achan->nsamp = cd1_nsamp + 2;
	if (cd1_length < 0 || cd1_length > achan->nsamp * 4 + 40)
	    return -1;
	RESIZE_BUFFER (int *, achan->data, achan->datasz, achan->nsamp + 1);
	datsz = achan->nsamp * sizeof (int);
	achan->nsamp -= 2;
	memcpy (&cd1_status, cp, 4);
	cp += 1 * 4;
	retcode = 0;
	if (achan->nsamp > 0) {
	    if (achan->iuser1) {
		nb = cd1_length - 16 - nauth;
		retcode = uncanada (achan->data, achan->nsamp, datsz, cp, nb);
		cp += nb;
	    } else {
		nb = cd1_length - 16 - nauth;
		if (nb == 3 * achan->nsamp) {
		    N32H4 (achan->data, cp, achan->nsamp);
		} else if (nb == 4 * achan->nsamp) {
		    N2H4 ((char *) achan->data, cp, achan->nsamp);
		} else if (nb == 2 * achan->nsamp) {
		    NS2HI (achan->data, cp, achan->nsamp);
		} else {
		    return -1;
		};
		cp += nb;
		achan->iuser3 = nb;
	    };
	};
	strncpy (achan->net, pkt->parts.src_net, PKT_NAMESIZE);
	achan->cuser1[0] = cd1_status[0];
	achan->cuser1[1] = cd1_status[1];
	achan->cuser1[2] = cd1_status[2];
	achan->cuser1[3] = cd1_status[3];
    };
    if (nbytes < cp - pp) {
	complain (0, "nbytes=%d < cp-pp=%d\n", nbytes, cp - pp);
	retcode = -1;
    };
    return retcode;
}


int
stuff_GCF (Packet * pkt, char *srcname, double *opkttime, char **ppp, int *nbytes, int *ppsz)
{
    char           *pp,
                   *cp;
    int             retcode = 0;
    unsigned int    sysid;
    unsigned int    streamid;
    unsigned int    datecode;
    char            sysid_str[7],
                    streamid_str[7];
    int            *gcf_diffs = 0;
    int             nrec = 0;
    int             ric;
    int             i;
    int             gcfversion;
    int             ICHAN = 0;
    PktChannel     *achan;
    achan = gettbl (pkt->channels, 0);
    pp = cp = *ppp;
    RESIZE_BUFFER (char *, *ppp, *ppsz, 1063 + 12);
    cp = *ppp + (cp - pp);
    pp = *ppp;
    if (pkt->version > 1) {
	return -1;
    }
    if (achan->iuser1 != GCF_MOTOROLA_BYTEORDER && achan->iuser1 != GCF_INTEL_BYTEORDER) {
	achan->iuser1 = GCF_MOTOROLA_BYTEORDER;
    };
    cp += 1060;
    HI2NUC (cp, &achan->iuser1, 1);
    cp += 1 * 1;
    cp -= 1061;
    if (achan->iuser1 == GCF_MOTOROLA_BYTEORDER) {
	cp += 1;
	HD2NUC (cp, &achan->samprate, 1);
	cp += 1 * 1;
	RESIZE_BUFFER (int *, achan->data, achan->datasz, achan->nsamp + 1);
	switch (achan->iuser2) {
	  case 1:
	    H2N4 (cp, (char *) gcf_diffs, achan->nsamp);
	    cp += 4 * achan->nsamp;
	    break;
	  case 2:
	    HI2NS (cp, (int *) gcf_diffs, achan->nsamp);
	    cp += 2 * achan->nsamp;
	    break;
	  case 4:
	    HI2NC (cp, gcf_diffs, achan->nsamp);
	    cp += 1 * achan->nsamp;
	    break;
	}
	cp += (GCF_MAX_NREC - nrec) * sizeof (int);
	HI2NUS (cp, &achan->iuser3, 1);
	cp += 2 * 1;
    } else if (achan->iuser1 == GCF_INTEL_BYTEORDER) {
	cp += 1;
	HD2NUC (cp, &achan->samprate, 1);
	cp += 1 * 1;
	RESIZE_BUFFER (int *, achan->data, achan->datasz, achan->nsamp + 1);
	switch (achan->iuser2) {
	  case 1:
	    HI2II (cp, gcf_diffs, achan->nsamp);
	    cp += 4 * achan->nsamp;
	    break;
	  case 2:
	    HI2IS (cp, gcf_diffs, achan->nsamp);
	    cp += 2 * achan->nsamp;
	    break;
	  case 4:
	    HI2NC (cp, gcf_diffs, achan->nsamp);
	    cp += 1 * achan->nsamp;
	    break;
	}
	cp += (GCF_MAX_NREC - nrec) * sizeof (int);
	HI2IUS (cp, &achan->iuser3, 1);
	cp += 2 * 1;
    } else {
	complain (0, "gcf: unknown byte order\n");
	retcode = -1;
    };
    *nbytes = cp - pp;
    insist (*nbytes < 1063 + 12);
    return retcode;
}



int
unstuff_GCF (char *srcname, double ipkttime, char *packet, int nbytes, Packet * pkt)
{
    char           *pp,
                   *cp;
    int             retcode = 0;
    unsigned int    sysid;
    unsigned int    streamid;
    unsigned int    datecode;
    char            sysid_str[7],
                    streamid_str[7];
    int            *gcf_diffs = 0;
    int             nrec = 0;
    int             ric;
    int             i;
    int             gcfversion;
    int             ICHAN = 0;
    PktChannel     *achan;

    achan = gettbl (pkt->channels, 0);
    if (achan == 0) {
	achan = newPktChannel ();
	pushtbl (pkt->channels, achan);
    }
    pp = cp = packet;
    if (nbytes == 1063) {
	pkt->version = 0;
	strncpy (achan->segtype, "V", 4);
	achan->calib = 0;
	achan->calper = 0;
    } else {
	NS2HI (&pkt->version, cp, 1);
	cp += 2 * 1;
	memcpy (&achan->segtype, cp, 2);
	cp += 1 * 2;
	NF2HD (&achan->calib, cp, 1);
	cp += 4 * 1;
	NF2HD (&achan->calper, cp, 1);
	cp += 4 * 1;
    };
    if (pkt->version > 1) {
	return -1;
    }
    pkt->nchannels = 1;
    strncpy (achan->net, pkt->parts.src_net, PKT_NAMESIZE);
    strncpy (achan->sta, pkt->parts.src_sta, PKT_NAMESIZE);
    strncpy (achan->chan, pkt->parts.src_chan, PKT_NAMESIZE);
    cp += 1060;
    NUC2HI (&achan->iuser1, cp, 1);
    cp += 1 * 1;
    cp -= 1061;
    if (achan->iuser1 == GCF_MOTOROLA_BYTEORDER) {
	N2H4 ((char *) &sysid, cp, 1);
	cp += 4 * 1;
	decodebase36 (sysid, sysid_str, 6);
	N2H4 ((char *) &streamid, cp, 1);
	cp += 4 * 1;
	decodebase36 (streamid, streamid_str, 6);
	sprintf (achan->cuser2, "%s_%s", sysid_str, streamid_str);
	N2H4 ((char *) &datecode, cp, 1);
	cp += 4 * 1;
	achan->time = datecode2epoch (datecode);
	cp += 1;
	NUC2HD (&achan->samprate, cp, 1);
	cp += 1 * 1;
	NUC2HI (&achan->iuser2, cp, 1);
	cp += 1 * 1;
	NUC2HI (&nrec, cp, 1);
	cp += 1 * 1;
	achan->nsamp = nrec * achan->iuser2;
	if (achan->nsamp < 0 || achan->nsamp > MAX_NSAMP)
	    return -1;
	allot (int *, gcf_diffs, achan->nsamp);
	RESIZE_BUFFER (int *, achan->data, achan->datasz, achan->nsamp + 1);
	N2H4 ((char *) achan->data, cp, 1);
	cp += 4 * 1;
	switch (achan->iuser2) {
	  case 1:
	    N2H4 ((char *) gcf_diffs, cp, achan->nsamp);
	    cp += 4 * achan->nsamp;
	    break;
	  case 2:
	    NS2HI (gcf_diffs, cp, achan->nsamp);
	    cp += 2 * achan->nsamp;
	    break;
	  case 4:
	    NC2HI (gcf_diffs, cp, achan->nsamp);
	    cp += 1 * achan->nsamp;
	    break;
	}
	N2H4 ((char *) &ric, cp, 1);
	cp += 4 * 1;
	cp += (GCF_MAX_NREC - nrec) * sizeof (int);
	for (i = 1; i < achan->nsamp; i++)
	    achan->data[i] = add24bit (achan->data[i - 1], gcf_diffs[i]);
	free (gcf_diffs);
	NUC2HI (&gcfversion, cp, 1);
	cp += 1 * 1;
	memcpy (achan->cuser1, cp, 33);
	cp += 1 * 33;
	achan->cuser1[32] = 0;
	NUS2HI (&achan->iuser3, cp, 1);
	cp += 2 * 1;
    } else if (achan->iuser1 == GCF_INTEL_BYTEORDER) {
	IUI2HUI (&sysid, cp, 1);
	cp += 4 * 1;
	decodebase36 (sysid, sysid_str, 6);
	IUI2HUI (&streamid, cp, 1);
	cp += 4 * 1;
	decodebase36 (streamid, streamid_str, 6);
	sprintf (achan->cuser2, "%s_%s", sysid_str, streamid_str);
	IUI2HUI (&datecode, cp, 1);
	cp += 4 * 1;
	achan->time = datecode2epoch (datecode);
	cp += 1;
	NUC2HD (&achan->samprate, cp, 1);
	cp += 1 * 1;
	NUC2HI (&achan->iuser2, cp, 1);
	cp += 1 * 1;
	NUC2HI (&nrec, cp, 1);
	cp += 1 * 1;
	achan->nsamp = nrec * achan->iuser2;
	if (achan->nsamp < 0 || achan->nsamp > MAX_NSAMP)
	    return -1;
	allot (int *, gcf_diffs, achan->nsamp);
	RESIZE_BUFFER (int *, achan->data, achan->datasz, achan->nsamp + 1);
	II2HI (achan->data, cp, 1);
	cp += 4 * 1;
	switch (achan->iuser2) {
	  case 1:
	    II2HI (gcf_diffs, cp, achan->nsamp);
	    cp += 4 * achan->nsamp;
	    break;
	  case 2:
	    IS2HI (gcf_diffs, cp, achan->nsamp);
	    cp += 2 * achan->nsamp;
	    break;
	  case 4:
	    NC2HI (gcf_diffs, cp, achan->nsamp);
	    cp += 1 * achan->nsamp;
	    break;
	}
	II2HI (&ric, cp, 1);
	cp += 4 * 1;
	cp += (GCF_MAX_NREC - nrec) * sizeof (int);
	for (i = 1; i < achan->nsamp; i++)
	    achan->data[i] = add24bit (achan->data[i - 1], gcf_diffs[i]);
	free (gcf_diffs);
	NUC2HI (&gcfversion, cp, 1);
	cp += 1 * 1;
	memcpy (achan->cuser1, cp, 33);
	cp += 1 * 33;
	achan->cuser1[32] = 0;
	IUS2HI (&achan->iuser3, cp, 1);
	cp += 2 * 1;
    } else {
	complain (0, "gcf: unknown byte order\n");
	retcode = -1;
    };
    if (nbytes < cp - pp) {
	complain (0, "nbytes=%d < cp-pp=%d\n", nbytes, cp - pp);
	retcode = -1;
    };
    return retcode;
}


int
stuff_GCFS (Packet * pkt, char *srcname, double *opkttime, char **ppp, int *nbytes, int *ppsz)
{
    char           *pp,
                   *cp;
    int             retcode = 0;
    int             nrecs;
    pp = cp = *ppp;
    RESIZE_BUFFER (char *, *ppp, *ppsz, 1063 + 2);
    cp = *ppp + (cp - pp);
    pp = *ppp;
    if (pkt->version > 1) {
	return -1;
    }
    cp += 15;
    HI2NUC (cp, &nrecs, 1);
    cp += 1 * 1;
    pkt->string[nrecs * 4] = 0;
    *nbytes = cp - pp;
    insist (*nbytes < 1063 + 2);
    return retcode;
}



int
unstuff_GCFS (char *srcname, double ipkttime, char *packet, int nbytes, Packet * pkt)
{
    char           *pp,
                   *cp;
    int             retcode = 0;
    int             nrecs;
    pp = cp = packet;
    if (nbytes == 1063) {
	pkt->version = 0;
    } else {
	NS2HI (&pkt->version, cp, 1);
	cp += 2 * 1;
    };
    if (pkt->version > 1) {
	return -1;
    }
    cp += 15;
    NUC2HI (&nrecs, cp, 1);
    cp += 1 * 1;
    allot (char *, pkt->string, 1008);
    strncpy ((char *) pkt->string, (char *) cp, 1008);
    cp += MIN (strlen (pkt->string) + 1, 1008);
    pkt->string[nrecs * 4] = 0;
    if (nbytes < cp - pp) {
	complain (0, "nbytes=%d < cp-pp=%d\n", nbytes, cp - pp);
	retcode = -1;
    };
    return retcode;
}


int
stuff_GEN (Packet * pkt, char *srcname, double *opkttime, char **ppp, int *nbytes, int *ppsz)
{
    char           *pp,
                   *cp;
    int             retcode = 0;
    unsigned char  *buf = 0;
    int             bufsz = 0;
    int             bytecount;
    int             i,
                   *idata;
    int             ICHAN = 0;
    PktChannel     *achan;
    achan = gettbl (pkt->channels, 0);
    pp = cp = *ppp;
    RESIZE_BUFFER (char *, *ppp, *ppsz, 48 + 5 * achan->nsamp);
    cp = *ppp + (cp - pp);
    pp = *ppp;
    HD2NF (cp, &achan->samprate, 1);
    cp += 4 * 1;
    HD2NF (cp, &achan->calib, 1);
    cp += 4 * 1;
    HD2NF (cp, &achan->calper, 1);
    cp += 4 * 1;
    HI2NS (cp, (int *) &achan->nsamp, 1);
    cp += 2 * 1;
    memcpy (cp, &achan->segtype, 2);
    cp += 1 * 2;
    *opkttime = achan->time;
    bytecount = cp - pp;
    H2N4 (cp, (char *) achan->data, achan->nsamp);
    cp += achan->nsamp * sizeof (int);
    *nbytes = cp - pp;
    insist (*nbytes < 48 + 5 * achan->nsamp);
    return retcode;
}



int
unstuff_GEN (char *srcname, double ipkttime, char *packet, int nbytes, Packet * pkt)
{
    char           *pp,
                   *cp;
    int             retcode = 0;
    unsigned char  *buf = 0;
    int             bufsz = 0;
    int             bytecount;
    int             i,
                   *idata;
    int             ICHAN = 0;
    PktChannel     *achan;

    achan = gettbl (pkt->channels, 0);
    if (achan == 0) {
	achan = newPktChannel ();
	pushtbl (pkt->channels, achan);
    }
    pp = cp = packet;
    pkt->nchannels = 1;
    NF2HD (&achan->samprate, cp, 1);
    cp += 4 * 1;
    NF2HD (&achan->calib, cp, 1);
    cp += 4 * 1;
    NF2HD (&achan->calper, cp, 1);
    cp += 4 * 1;
    NS2HI (&achan->nsamp, cp, 1);
    cp += 2 * 1;
    if (achan->nsamp < 0 || achan->nsamp > MAX_NSAMP)
	return -1;
    memcpy (&achan->segtype, cp, 2);
    cp += 1 * 2;
    achan->segtype[2] = 0;
    strncpy (achan->net, pkt->parts.src_net, PKT_NAMESIZE);
    strncpy (achan->sta, pkt->parts.src_sta, PKT_NAMESIZE);
    strncpy (achan->chan, pkt->parts.src_chan, PKT_NAMESIZE);
    strncpy (achan->loc, pkt->parts.src_loc, PKT_NAMESIZE);
    achan->time = ipkttime;
    RESIZE_BUFFER (int *, achan->data, achan->datasz, achan->nsamp + 1);
    N2H4 ((char *) achan->data, cp, achan->nsamp);
    cp += achan->nsamp * sizeof (int);
    if (nbytes < cp - pp) {
	complain (0, "nbytes=%d < cp-pp=%d\n", nbytes, cp - pp);
	retcode = -1;
    };
    return retcode;
}


int
stuff_GENC (Packet * pkt, char *srcname, double *opkttime, char **ppp, int *nbytes, int *ppsz)
{
    char           *pp,
                   *cp;
    int             retcode = 0;
    unsigned char  *buf = 0;
    int             bufsz = 0;
    int             bytecount;
    int             i,
                   *idata;
    int             ICHAN = 0;
    PktChannel     *achan;
    achan = gettbl (pkt->channels, 0);
    pp = cp = *ppp;
    RESIZE_BUFFER (char *, *ppp, *ppsz, 48 + 5 * achan->nsamp);
    cp = *ppp + (cp - pp);
    pp = *ppp;
    HD2NF (cp, &achan->samprate, 1);
    cp += 4 * 1;
    HD2NF (cp, &achan->calib, 1);
    cp += 4 * 1;
    HD2NF (cp, &achan->calper, 1);
    cp += 4 * 1;
    HI2NS (cp, (int *) &achan->nsamp, 1);
    cp += 2 * 1;
    memcpy (cp, &achan->segtype, 2);
    cp += 1 * 2;
    strncpy (pkt->parts.src_net, achan->net, PKT_NAMESIZE);
    strncpy (pkt->parts.src_sta, achan->sta, PKT_NAMESIZE);
    strncpy (pkt->parts.src_chan, achan->chan, PKT_NAMESIZE);
    strncpy (pkt->parts.src_loc, achan->loc, PKT_NAMESIZE);
    *opkttime = achan->time;
    bytecount = cp - pp;
    allot (int *, idata, achan->nsamp);
    idata[0] = achan->data[0];
    for (i = 1; i < achan->nsamp; i++)
	idata[i] = achan->data[i] - achan->data[i - 1];
    buf = 0;
    bufsz = 0;
    retcode = gencompress ((unsigned char **) &buf, &bytecount, &bufsz, idata, achan->nsamp, 0);
    free (idata);
    memcpy (cp, buf, bytecount);
    free (buf);
    cp += bytecount;
    *nbytes = cp - pp;
    insist (*nbytes < 48 + 5 * achan->nsamp);
    return retcode;
}



int
unstuff_GENC (char *srcname, double ipkttime, char *packet, int nbytes, Packet * pkt)
{
    char           *pp,
                   *cp;
    int             retcode = 0;
    unsigned char  *buf = 0;
    int             bufsz = 0;
    int             bytecount;
    int             i,
                   *idata;
    int             ICHAN = 0;
    PktChannel     *achan;

    achan = gettbl (pkt->channels, 0);
    if (achan == 0) {
	achan = newPktChannel ();
	pushtbl (pkt->channels, achan);
    }
    pp = cp = packet;
    pkt->nchannels = 1;
    NF2HD (&achan->samprate, cp, 1);
    cp += 4 * 1;
    NF2HD (&achan->calib, cp, 1);
    cp += 4 * 1;
    NF2HD (&achan->calper, cp, 1);
    cp += 4 * 1;
    NS2HI (&achan->nsamp, cp, 1);
    cp += 2 * 1;
    if (achan->nsamp < 0 || achan->nsamp > MAX_NSAMP)
	return -1;
    memcpy (&achan->segtype, cp, 2);
    cp += 1 * 2;
    achan->segtype[2] = 0;
    strncpy (achan->net, pkt->parts.src_net, PKT_NAMESIZE);
    strncpy (achan->sta, pkt->parts.src_sta, PKT_NAMESIZE);
    strncpy (achan->chan, pkt->parts.src_chan, PKT_NAMESIZE);
    strncpy (achan->loc, pkt->parts.src_loc, PKT_NAMESIZE);
    achan->time = ipkttime;
    RESIZE_BUFFER (int *, achan->data, achan->datasz, achan->nsamp + 1);
    achan->datasz *= sizeof (int);
    retcode = genuncompress (&achan->data, &achan->nsamp, &achan->datasz, (unsigned char *) cp, nbytes - (cp - pp));
    achan->datasz /= sizeof (int);
    for (i = 1; i < achan->nsamp; i++)
	achan->data[i] += achan->data[i - 1];
    if (nbytes < cp - pp) {
	complain (0, "nbytes=%d < cp-pp=%d\n", nbytes, cp - pp);
	retcode = -1;
    };
    return retcode;
}


int
stuff_MGENC (Packet * pkt, char *srcname, double *opkttime, char **ppp, int *nbytes, int *ppsz)
{
    char           *pp,
                   *cp;
    int             retcode = 0;
    unsigned char  *buf = 0;
    int             bufsz = 0;
    int             bytecount;
    int             i,
                   *idata;
    int             ICHAN = 0;
    PktChannel     *achan;
    achan = gettbl (pkt->channels, 0);
    pp = cp = *ppp;
    RESIZE_BUFFER (char *, *ppp, *ppsz, (4 + (50 + 2000 * 4) * pkt->nchannels));
    cp = *ppp + (cp - pp);
    pp = *ppp;
    pkt->version = 2;
    HI2NS (cp, (int *) &pkt->version, 1);
    cp += 2 * 1;
    HI2NS (cp, (int *) &pkt->nchannels, 1);
    cp += 2 * 1;
    for (ICHAN = 0; ICHAN < pkt->nchannels; ICHAN++) {
	achan = gettbl (pkt->channels, ICHAN);
	if (achan == 0) {
	    achan = newPktChannel ();
	    pushtbl (pkt->channels, achan);
	};
	if (pkt->version == 1) {
	    if (pkt->nchannels == 1) {
		strncpy (pkt->parts.src_net, achan->net, PKT_NAMESIZE);
		strncpy (pkt->parts.src_sta, achan->sta, PKT_NAMESIZE);
		strncpy (pkt->parts.src_chan, achan->chan, PKT_NAMESIZE);
		strncpy (pkt->parts.src_loc, achan->loc, PKT_NAMESIZE);
		*opkttime = achan->time;
	    } else {
		memcpy (cp, &achan->net, PKT_NAMESIZE);
		cp += 1 * PKT_NAMESIZE;
		memcpy (cp, &achan->sta, PKT_NAMESIZE);
		cp += 1 * PKT_NAMESIZE;
		if (ICHAN == 0) {
		    strncpy (pkt->parts.src_net, achan->net, PKT_NAMESIZE);
		    strncpy (pkt->parts.src_sta, achan->sta, PKT_NAMESIZE);
		};
		memcpy (cp, &achan->chan, PKT_NAMESIZE);
		cp += 1 * PKT_NAMESIZE;
		memcpy (cp, &achan->loc, PKT_NAMESIZE);
		cp += 1 * PKT_NAMESIZE;
	    };
	} else {
	    memcpy (cp, &achan->net, PKT_NAMESIZE);
	    cp += 1 * PKT_NAMESIZE;
	    memcpy (cp, &achan->sta, PKT_NAMESIZE);
	    cp += 1 * PKT_NAMESIZE;
	    if (ICHAN == 0) {
		strncpy (pkt->parts.src_net, achan->net, PKT_NAMESIZE);
		strncpy (pkt->parts.src_sta, achan->sta, PKT_NAMESIZE);
	    };
	    memcpy (cp, &achan->chan, PKT_NAMESIZE);
	    cp += 1 * PKT_NAMESIZE;
	    memcpy (cp, &achan->loc, PKT_NAMESIZE);
	    cp += 1 * PKT_NAMESIZE;
	};
	memcpy (cp, &achan->segtype, 2);
	cp += 1 * 2;
	if (pkt->version == 1) {
	    if (pkt->nchannels > 1) {
		H2ND ((uchar *) cp, &achan->time, 1);
		cp += 8 * 1;
		if (ICHAN == 0) {
		    *opkttime = achan->time;
		};
	    };
	} else {
	    H2ND ((uchar *) cp, &achan->time, 1);
	    cp += 8 * 1;
	    if (ICHAN == 0) {
		*opkttime = achan->time;
	    };
	};
	HI2NS (cp, (int *) &achan->nsamp, 1);
	cp += 2 * 1;
	HD2NF (cp, &achan->samprate, 1);
	cp += 4 * 1;
	HD2NF (cp, &achan->calib, 1);
	cp += 4 * 1;
	HD2NF (cp, &achan->calper, 1);
	cp += 4 * 1;
	if (achan->nsamp < 5) {
	    H2N4 (cp, (char *) achan->data, achan->nsamp);
	    cp += 4 * achan->nsamp;
	    bytecount = achan->nsamp * 4;
	} else {
	    allot (int *, idata, achan->nsamp);
	    idata[0] = achan->data[0];
	    for (i = 1; i < achan->nsamp; i++)
		idata[i] = achan->data[i] - achan->data[i - 1];
	    buf = 0;
	    bufsz = 0;
	    retcode = gencompress ((unsigned char **) &buf, &bytecount, &bufsz, idata, achan->nsamp, 0);
	    free (idata);
	    HI2NS (cp, (int *) &bytecount, 1);
	    cp += 2 * 1;
	    memcpy (cp, buf, bytecount);
	    free (buf);
	};
	cp += bytecount;
    };
    *nbytes = cp - pp;
    insist (*nbytes < (4 + (50 + 2000 * 4) * pkt->nchannels));
    return retcode;
}



int
unstuff_MGENC (char *srcname, double ipkttime, char *packet, int nbytes, Packet * pkt)
{
    char           *pp,
                   *cp;
    int             retcode = 0;
    unsigned char  *buf = 0;
    int             bufsz = 0;
    int             bytecount;
    int             i,
                   *idata;
    int             ICHAN = 0;
    PktChannel     *achan;

    achan = gettbl (pkt->channels, 0);
    if (achan == 0) {
	achan = newPktChannel ();
	pushtbl (pkt->channels, achan);
    }
    pp = cp = packet;
    NS2HI (&pkt->version, cp, 1);
    cp += 2 * 1;
    NS2HI (&pkt->nchannels, cp, 1);
    cp += 2 * 1;
    for (ICHAN = 0; ICHAN < pkt->nchannels; ICHAN++) {
	achan = gettbl (pkt->channels, ICHAN);
	if (achan == 0) {
	    achan = newPktChannel ();
	    pushtbl (pkt->channels, achan);
	};
	if (pkt->version == 1) {
	    if (pkt->nchannels == 1) {
		strncpy (achan->net, pkt->parts.src_net, PKT_NAMESIZE);
		strncpy (achan->sta, pkt->parts.src_sta, PKT_NAMESIZE);
		strncpy (achan->chan, pkt->parts.src_chan, PKT_NAMESIZE);
		strncpy (achan->loc, pkt->parts.src_loc, PKT_NAMESIZE);
		achan->time = ipkttime;
	    } else {
		memcpy (&achan->net, cp, PKT_NAMESIZE);
		cp += 1 * PKT_NAMESIZE;
		memcpy (&achan->sta, cp, PKT_NAMESIZE);
		cp += 1 * PKT_NAMESIZE;
		memcpy (&achan->chan, cp, PKT_NAMESIZE);
		cp += 1 * PKT_NAMESIZE;
		memcpy (&achan->loc, cp, PKT_NAMESIZE);
		cp += 1 * PKT_NAMESIZE;
	    };
	} else {
	    memcpy (&achan->net, cp, PKT_NAMESIZE);
	    cp += 1 * PKT_NAMESIZE;
	    memcpy (&achan->sta, cp, PKT_NAMESIZE);
	    cp += 1 * PKT_NAMESIZE;
	    memcpy (&achan->chan, cp, PKT_NAMESIZE);
	    cp += 1 * PKT_NAMESIZE;
	    memcpy (&achan->loc, cp, PKT_NAMESIZE);
	    cp += 1 * PKT_NAMESIZE;
	};
	memcpy (&achan->segtype, cp, 2);
	cp += 1 * 2;
	achan->segtype[2] = 0;
	if (pkt->version == 1) {
	    if (pkt->nchannels > 1) {
		N2HD (&achan->time, (uchar *) cp, 1);
		cp += 8 * 1;
	    };
	} else {
	    N2HD (&achan->time, (uchar *) cp, 1);
	    cp += 8 * 1;
	};
	NS2HI (&achan->nsamp, cp, 1);
	cp += 2 * 1;
	if (achan->nsamp < 0 || achan->nsamp > MAX_NSAMP)
	    return -1;
	NF2HD (&achan->samprate, cp, 1);
	cp += 4 * 1;
	NF2HD (&achan->calib, cp, 1);
	cp += 4 * 1;
	NF2HD (&achan->calper, cp, 1);
	cp += 4 * 1;
	RESIZE_BUFFER (int *, achan->data, achan->datasz, achan->nsamp + 1);
	if (achan->nsamp < 5) {
	    N2H4 ((char *) achan->data, cp, achan->nsamp);
	    cp += 4 * achan->nsamp;
	    bytecount = achan->nsamp * 4;
	} else {
	    achan->datasz *= sizeof (int);
	    NS2HI (&bytecount, cp, 1);
	    cp += 2 * 1;
	    retcode = genuncompress (&achan->data, &achan->nsamp, &achan->datasz, (unsigned char *) cp, bytecount);
	    achan->datasz /= sizeof (int);
	    for (i = 1; i < achan->nsamp; i++)
		achan->data[i] += achan->data[i - 1];
	};
	cp += bytecount;
    };
    if (nbytes < cp - pp) {
	complain (0, "nbytes=%d < cp-pp=%d\n", nbytes, cp - pp);
	retcode = -1;
    };
    return retcode;
}


int
stuff_pf (Packet * pkt, char *srcname, double *opkttime, char **ppp, int *nbytes, int *ppsz)
{
    char           *pp,
                   *cp;
    int             retcode = 0;
    char           *str;
    int             n;
    pp = cp = *ppp;
    str = pf2string (pkt->pf);
    pkt->version = 1;
    n = strlen (str);
    RESIZE_BUFFER (char *, *ppp, *ppsz, n + 10);
    cp = *ppp + (cp - pp);
    pp = *ppp;
    HI2NS (cp, (int *) &pkt->version, 1);
    cp += 2 * 1;
    switch (pkt->version) {
      case 0:
	HI2NS (cp, (int *) &Zero, 1);
	cp += 2 * 1;
	H2N4 (cp, (char *) &Zero, 1);
	cp += 4 * 1;
	break;
      default:
	break;
    }
    *opkttime = now ();
    strncpy ((char *) cp, (char *) str, strlen (str) + 1);
    cp += MIN (strlen (str) + 1, strlen (str) + 1);
    free (str);
    *nbytes = cp - pp;
    insist (*nbytes < n + 10);
    return retcode;
}



int
unstuff_pf (char *srcname, double ipkttime, char *packet, int nbytes, Packet * pkt)
{
    char           *pp,
                   *cp;
    int             retcode = 0;
    pp = cp = packet;
    NS2HI (&pkt->version, cp, 1);
    cp += 2 * 1;
    switch (pkt->version) {
      case 0:
	cp += 6;
	nbytes -= 8;
	break;
      default:
	nbytes -= 2;
	break;
    }
    if (bad_ascii (cp, nbytes))
	return -1;
    pfcompile (cp, &pkt->pf);
    if (nbytes < cp - pp) {
	complain (0, "nbytes=%d < cp-pp=%d\n", nbytes, cp - pp);
	retcode = -1;
    };
    return retcode;
}


int
stuff_db (Packet * pkt, char *srcname, double *opkttime, char **ppp, int *nbytes, int *ppsz)
{
    char           *pp,
                   *cp;
    int             retcode = 0;
    char           *schema_name;
    char           *table_name;
    char            filename[FILENAME_MAX];
    int             mark;
    int             view,
                    record_size,
                    bytecount;
    int             fd,
                    nread;
    pp = cp = *ppp;
    RESIZE_BUFFER (char *, *ppp, *ppsz, 32767);
    cp = *ppp + (cp - pp);
    pp = *ppp;
    mutex_lock (&Database_mutex);
    dbquery (pkt->db, dbSCHEMA_NAME, &schema_name);
    dbquery (pkt->db, dbTABLE_NAME, &table_name);
    strcpy (pkt->parts.src_subcode, table_name);
    dbquery (pkt->db, dbVIEW_TABLE_COUNT, &view);
    if (view != 1) {
	register_error (0, "Table %s is not a primary table", table_name);
	retcode = -1;
    } else if (pkt->db.record < 0 && pkt->db.record != dbSCRATCH) {
	register_error (0, "a particular record must be specified");
	retcode = -1;
    } else {
	dbquery (pkt->db, dbRECORD_SIZE, &record_size);
	pkt->version = 1;
	*cp++ = 1;
	switch (pkt->version) {
	  case 0:
	    cp += 3;
	    HI2NS (cp, (int *) &pkt->pkttype->hdrcode, 1);
	    cp += 2 * 1;
	    HI2NS (cp, (int *) &pkt->pkttype->bodycode, 1);
	    cp += 2 * 1;
	    strcpy (cp, table_name);
	    cp += strlen (table_name);
	    *cp++ = ' ';
	    break;
	  case 1:
	    break;
	  default:
	    mutex_unlock (&Database_mutex);
	    retcode = -1;
	};
	strcpy (cp, schema_name);
	cp += strlen (schema_name);
	*cp++ = 0;
	if (pkt->version == 0) {
	    bytecount = cp - pp - 8;
	    HI2NS (pp, &bytecount, 1);
	};
	if (dbget (pkt->db, cp) < 0) {
	    retcode = -1;
	};
	cp += record_size;
	*cp++ = 0;
	mark = elog_mark ();
	if (dbextfile (pkt->db, table_name, filename) == 1) {
	    struct stat     statbuf;
	    stat (filename, &statbuf);
	    if (statbuf.st_size < 25000) {
		fd = open (filename, O_RDONLY);
		nread = read (fd, cp, statbuf.st_size);
		close (fd);
		cp += nread;
		*cp++ = 0;
	    };
	};
	elog_flush (0, mark);
	;
	if (pkt->version == 0) {
	    bytecount = record_size;
	    HI2NS (pp + 2, &bytecount, 1);
	};
	*opkttime = dbtime (pkt->db);
	mutex_unlock (&Database_mutex);
    };
    *nbytes = cp - pp;
    insist (*nbytes < 32767);
    return retcode;
}



int
unstuff_db (char *srcname, double ipkttime, char *packet, int nbytes, Packet * pkt)
{
    char           *pp,
                   *cp;
    int             retcode = 0;
    char           *p1,
                   *p2,
                    schema[FILENAME_MAX];
    int             n;
    int             view,
                    record_size,
                    bytecount;
    pp = cp = packet;
    NC2HI (&pkt->version, cp, 1);
    cp += 1 * 1;
    switch (pkt->version) {
      case 0:
	cp += 7;
	p1 = strchr (cp, ' ');
	if (p1 == 0)
	    return -1;
	else
	    p1++;
	p2 = strchr (cp, '\n');
	if (p2 == 0) {
	    p2 = p1 + strlen (p1);
	};
	n = MIN (p2 - p1, FILENAME_MAX);
	strncpy (schema, p1, n);
	schema[n] = 0;
	cp = p2 + 1;
	break;
      case 1:
	strncpy (schema, cp, FILENAME_MAX);
	cp += strlen (schema) + 1;
	break;
      default:
	return -1;
    };
    mutex_lock (&Database_mutex);
    pkt->db = temp_database (schema);
    if (pkt->db.database < 0) {
	mutex_unlock (&Database_mutex);
	return -1;
    };
    pkt->db = dblookup (pkt->db, 0, pkt->parts.src_subcode, 0, 0);
    if (pkt->db.table >= 0) {
	pkt->db.record = dbSCRATCH;
	retcode = dbput (pkt->db, cp);
	p1 = cp + strlen (cp) + 1;
	if (p1 < pp + nbytes) {
	    pkt->dfile = p1;
	    pkt->dfile_size = pp + nbytes - p1 - 1;
	} else {
	    pkt->dfile = 0;
	    pkt->dfile_size = 0;
	};
    } else {
	retcode = pkt->db.table;
    };
    mutex_unlock (&Database_mutex);
    if (nbytes < cp - pp) {
	complain (0, "nbytes=%d < cp-pp=%d\n", nbytes, cp - pp);
	retcode = -1;
    };
    return retcode;
}


int
stuff_ch (Packet * pkt, char *srcname, double *opkttime, char **ppp, int *nbytes, int *ppsz)
{
    char           *pp,
                   *cp;
    int             retcode = 0;
    int             size;
    pp = cp = *ppp;
    RESIZE_BUFFER (char *, *ppp, *ppsz, (pkt->string_size > 0 ? pkt->string_size : strlen (pkt->string)) + 10);
    cp = *ppp + (cp - pp);
    pp = *ppp;
    if (pkt->string_size == 0) {
	pkt->string_size = MIN (MAX_CH_SIZE, strlen (pkt->string));
    } else {
	pkt->string_size = MIN (MAX_CH_SIZE, pkt->string_size);
    };
    memcpy (cp, pkt->string, pkt->string_size);
    cp += pkt->string_size;
    *cp++ = 0;
    *opkttime = now ();
    *nbytes = cp - pp;
    insist (*nbytes < (pkt->string_size > 0 ? pkt->string_size : strlen (pkt->string)) + 10);
    return retcode;
}



int
unstuff_ch (char *srcname, double ipkttime, char *packet, int nbytes, Packet * pkt)
{
    char           *pp,
                   *cp;
    int             retcode = 0;
    int             size;
    pp = cp = packet;
    if (pkt->string)
	free (pkt->string);
    pkt->string_size = MIN (nbytes, MAX_CH_SIZE);
    allot (char *, pkt->string, pkt->string_size + 1);
    memcpy (pkt->string, cp, pkt->string_size);
    pkt->string[pkt->string_size] = 0;
    cp += pkt->string_size;
    if (nbytes < cp - pp) {
	complain (0, "nbytes=%d < cp-pp=%d\n", nbytes, cp - pp);
	retcode = -1;
    };
    return retcode;
}


int
stuff_elog (Packet * pkt, char *srcname, double *opkttime, char **ppp, int *nbytes, int *ppsz)
{
    return stuff_ch (pkt, srcname, opkttime, ppp, nbytes, ppsz);
}

int
unstuff_elog (char *srcname, double ipkttime, char *packet, int nbytes, Packet * pkt)
{
    return unstuff_ch (srcname, ipkttime, packet, nbytes, pkt);
}


int
stuff_syslog (Packet * pkt, char *srcname, double *opkttime, char **ppp, int *nbytes, int *ppsz)
{
    return stuff_ch (pkt, srcname, opkttime, ppp, nbytes, ppsz);
}

int
unstuff_syslog (char *srcname, double ipkttime, char *packet, int nbytes, Packet * pkt)
{
    return unstuff_ch (srcname, ipkttime, packet, nbytes, pkt);
}


int
stuff_log (Packet * pkt, char *srcname, double *opkttime, char **ppp, int *nbytes, int *ppsz)
{
    char           *pp,
                   *cp;
    int             retcode = 0;
    pp = cp = *ppp;
#define QUANTERRA_LOG_HDR_TYPE 0x716c
    RESIZE_BUFFER (char *, *ppp, *ppsz, strlen (pkt->string) + 20);
    cp = *ppp + (cp - pp);
    pp = *ppp;
    H2N2 (cp, (char *) &Zero, 1);
    cp += 2 * 1;
    H2N2 (cp, (char *) &Zero, 1);
    cp += 2 * 1;
    HI2NS (cp, (int *) &pkt->pkttype->hdrcode, 1);
    cp += 2 * 1;
    HI2NS (cp, (int *) &pkt->pkttype->bodycode, 1);
    cp += 2 * 1;
    strncpy ((char *) cp, (char *) pkt->string, 512);
    cp += MIN (strlen (pkt->string) + 1, 512);
    *nbytes = cp - pp;
    insist (*nbytes < strlen (pkt->string) + 20);
    return retcode;
}



int
unstuff_log (char *srcname, double ipkttime, char *packet, int nbytes, Packet * pkt)
{
    char           *pp,
                   *cp;
    int             retcode = 0;
    pp = cp = packet;
#define QUANTERRA_LOG_HDR_TYPE 0x716c
    cp += 6;
    NS2HI (&pkt->version, cp, 1);
    cp += 2 * 1;
    if (pkt->string)
	free (pkt->string);
    pkt->string = strdup (cp);
    pkt->string_size = strlen (pkt->string) + 1;
    cp += pkt->string_size;
    if (nbytes < cp - pp) {
	complain (0, "nbytes=%d < cp-pp=%d\n", nbytes, cp - pp);
	retcode = -1;
    };
    return retcode;
}


int
stuff_cmd (Packet * pkt, char *srcname, double *opkttime, char **ppp, int *nbytes, int *ppsz)
{
    return stuff_pf (pkt, srcname, opkttime, ppp, nbytes, ppsz);
}

int
unstuff_cmd (char *srcname, double ipkttime, char *packet, int nbytes, Packet * pkt)
{
    return unstuff_pf (srcname, ipkttime, packet, nbytes, pkt);
}


PacketType      PktList[] = {
    {"junk", "junk", Pkt_ch, 0, 0, "bad packets",
    stuff_junk, unstuff_junk},
    {"dbug", "dbug", Pkt_wf, 0, 0, "simple debugging packet type",
    stuff_dbug, unstuff_dbug},
    {"Cougar", "Cougar", Pkt_wf, 0xdeb, 0xdeb, "old type used by Cougar drum recorder",
    stuff_Cougar, unstuff_Cougar},
    {"iceworm", "", Pkt_wf, 0x0adb, 0xffffaadb, "Iceworm packets",
    stuff_iceworm, unstuff_iceworm},
    {"replay", "replay", Pkt_wf, 0, 0, "orb2db/dbreplay packets",
    stuff_replay, unstuff_replay},
    {"CPSCLHS", "CPSCLHS", Pkt_wf, 0x0bba, 0x4454, "100 sps data stream",
    stuff_CPSCLHS, unstuff_CPSCLHS},
    {"CPSCLLS", "CPSCLLS", Pkt_wf, 0x0bba, 0x4454, "40 sps data stream",
    stuff_CPSCLLS, unstuff_CPSCLLS},
    {"BBA", "BBA", Pkt_wf, 0x0bba, 0xffffdaab, "Broadband array data stream 5/18/01",
    stuff_BBA, unstuff_BBA},
    {"CBBHS", "CBBHS", Pkt_wf, 0x0bba, 0xffffbbcd, "100 sps data stream",
    stuff_CBBHS, unstuff_CBBHS},
    {"CBBLS", "CBBLS", Pkt_wf, 0x0bba, 0xffffbbcd, "40 sps data stream",
    stuff_CBBLS, unstuff_CBBLS},
    {"CBB1S", "CBB1S", Pkt_wf, 0x0bba, 0xffffbbcd, "1 sps data stream",
    stuff_CBB1S, unstuff_CBB1S},
    {"CBB1S_delayed", "CBB1S_delayed", Pkt_wf, 0x0bba, 0xffffbbcd, "1 sps data stream delayed",
    stuff_CBB1S_delayed, unstuff_CBB1S_delayed},
    {"CALS", "CALS", Pkt_wf, 0x0bba, 0xffffbbcd, "40 sps data stream",
    stuff_CALS, unstuff_CALS},
    {"CAHS", "CAHS", Pkt_wf, 0x0bba, 0xffffbbcd, "100 sps data stream",
    stuff_CAHS, unstuff_CAHS},
    {"B3S2DP", "B3S2DP", Pkt_wf, 0x0bba, 0xffffbbcd, "",
    stuff_B3S2DP, unstuff_B3S2DP},
    {"BSP", "BSP", Pkt_st, 0, 0, "data concentrator status packet",
    stuff_BSP, unstuff_BSP},
    {"LISS", "LISS", Pkt_wf, 0x6c6c, 0x6c6c, "wrapped SEED packets",
    stuff_LISS, unstuff_LISS},
    {"SEED", "SEED", Pkt_wf, 0, 0, "wrapped SEED packets without the confusing old header stuff",
    stuff_SEED, unstuff_SEED},
    {"MSEED", "MSEED", Pkt_wf, 0, 0, "wrapped SEED packet, ala Neuhauser, lacking calib/calper/segtype and commingling logs",
    stuff_MSEED, unstuff_MSEED},
    {"raw", "raw", Pkt_rw, 0, 0, "binary data wrapped into orb packet",
    stuff_raw, unstuff_raw},
    {"file", "file", Pkt_rw, 0, 0, "binary data wrapped into orb packet",
    stuff_file, unstuff_file},
    {"xfer", "xfer", Pkt_fl, 0, 0, "file fragments from orbxfer2(1)",
    stuff_xfer, unstuff_xfer},
    {"QCDAT", "QCDAT", Pkt_wf, 0x7165, 0x7165, "data from qt2orb(1)",
    stuff_QCDAT, unstuff_QCDAT},
    {"CD1S", "CD1S", Pkt_wf, 0, 0, "CD-1 Channel sub-frame packet",
    stuff_CD1S, unstuff_CD1S},
    {"CD1", "CD1", Pkt_wf, 0, 0, "CD-1 data frame packet",
    stuff_CD1, unstuff_CD1},
    {"GCF", "GCF", Pkt_wf, 0, 0, "Guralp Compressed Format",
    stuff_GCF, unstuff_GCF},
    {"GCFS", "GCFS", Pkt_st, 0, 0, "Guralp Compressed Format Status packet",
    stuff_GCFS, unstuff_GCFS},
    {"GEN", "GEN", Pkt_wf, 0, 0, "Generic data frame packet",
    stuff_GEN, unstuff_GEN},
    {"GENC", "GENC", Pkt_wf, 0, 0, "Generic compressed data frame packet",
    stuff_GENC, unstuff_GENC},
    {"MGENC", "MGENC", Pkt_wf, 0, 0, "Multiplexed generic compressed data frame packet",
    stuff_MGENC, unstuff_MGENC},
    {"pf", "pf", Pkt_pf, 3, 3, "parameter file",
    stuff_pf, unstuff_pf},
    {"db", "db", Pkt_db, 2, 2, "database record",
    stuff_db, unstuff_db},
    {"ch", "ch", Pkt_ch, 0, 0, "character string",
    stuff_ch, unstuff_ch},
    {"elog", "elog", Pkt_ch, 0, 0, "elog error messages",
    stuff_elog, unstuff_elog},
    {"syslog", "syslog", Pkt_ch, 0, 0, "system log messages",
    stuff_syslog, unstuff_syslog},
    {"log", "log", Pkt_ch, 0x716c, 0x716c, "Quanterra log messages",
    stuff_log, unstuff_log},
    {"cmd", "cmd", Pkt_cn, 0, 0, "Command & Control messages",
    stuff_cmd, unstuff_cmd},
};
int             nPktList = sizeof (PktList) / sizeof (PacketType);
