#include "coords.h"
#include "stock.h"
#include "Pkt.h"

static void
show_stash ( FILE *file, Packet *pkt)
{
    fprintf ( file, "%s\n", pkt->string ) ; 
}

static void
show_fl (int pktid, char *srcname, double pkttime, char *packet, int nbytes, Packet *unstuffed, FILE *file, int mode ) 
{
    char *id, *cp ;
    int i, seq ; 
    id = strdup ( unstuffed->parts.src_subcode ) ; 
    cp = strchr(id, '/') ; 
    if ( cp != 0 ) { 
	*cp = 0 ; 
    }
    fprintf(file, "    id       = %s\n", id ) ; 
    seq = unstuffed->string_size ;
    fprintf(file, "    sequence = %8d\n", seq ) ; 
    if ( seq == 0 ) { 
	fprintf(file, "    filename   = %s\n", unstuffed->dfile ) ; 
	fprintf(file, "    mode       = %8o\n", unstuffed->nchannels ) ;
    }
    fprintf(file, "    digest   = " ) ; 
    for (i = 0; i < 16; i++) {
        printf ("%02x", unstuffed->string[i] & 0xff );
    }
    if ( unstuffed->dfile_size == 0 ) { 
	fprintf(file, "    *trailer packet*\n" ) ; 
    } else { 
	fprintf(file, "    ndata    = %8d\n", unstuffed->dfile_size ) ; 
    }
}

static void
show_db ( FILE *file, Dbptr db )  
{
    char *table_name, *field_name ; 
    char field[STRSZ], null[STRSZ] ; 
    int nfields ; 
    Dbptr db2 ;

    dbquery ( db, dbTABLE_NAME, &table_name ) ;
    fprintf ( file, "    Table %s\n", table_name ) ;
    dbquery ( db, dbFIELD_COUNT, &nfields ) ; 
    for ( db.field = 0 ; db.field < nfields ; db.field++ ) {
	dbget ( db, field ) ; 
	db2 = db ; db2.record = dbNULL ;
	dbget ( db2, null ) ;
	if ( strcmp ( field, null ) != 0 ) { 
	    dbquery ( db, dbFIELD_NAME, &field_name ) ; 
	    fprintf ( file, "    %s = '%s'\n", field_name, field ) ; 
	}
    }
}

static void
showPktSpecific(FILE *file, Packet *pkt) 
{
    if ( strncmp(pkt->parts.src_suffix, "CD1", 3) == 0 ) { 
	if ( pkt->string != 0 ) { 
	    fprintf(file, "weather: %s\n", pkt->string ) ;
	}
    }
}

static char *setStatus[] = {
    "signature at digitizer",		/*  0 */
    "**error**",			/*  1 */
    "**error**",			/*  2 */
    "**error**",			/*  3 */
    "**error**",			/*  4 */
    "**error**",			/*  5 */
    "**error**",			/*  6 */
    "**error**",			/*  7 */
    "**error**",			/*  8 */
    "**error**",			/*  9 */
    0,					/* 10 */
    0,					/* 11 */
    0,					/* 12 */
    0,					/* 13 */
    0,					/* 14 */
    0,					/* 15 */
    "digitizer calibrator loopback on",	/* 16 */
    "digitizer analog input shorted",	/* 17 */
    "Intrasite communications buffer overflow",		/* 18 */
    "GPS antenna or receiver off",	/* 19 */
    "clock differential too large", 	/* 20 */
    "equipment moved", 			/* 21 */
    "authentication unit opened", 	/* 22 */
    "vault door open",			/* 23 */
    "**error**",			/* 24 */
    "**error**",			/* 25 */
    0,					/* 26 */
    0,					/* 27 */
    "calibration signal", 		/* 28 */
    "data clipped by digitizer", 			/* 29 */
    "zeroed data from digitizer", 		/* 30 */
    "dead channel from digitizer",		/* 31 */
} ;

static void
show_CD1S_status ( FILE *file, char *status ) 
{
    int stat, i, bit, voltage ;
    memcpy ( &stat, status, 4 ) ;
    for ( i=31 ; i>=0 ; i-- ) { 
	bit = 1 << i ; 
	if ( bit & stat ) { 
	    if ( setStatus[i] != 0 ) { 
		fprintf ( file, "bit #%2d: %s\n", i, setStatus[i] ) ;
	    }
	} 
    }
    if ( stat & (1 <<28) ) { 
	switch ( (stat>>26) & 3 ) { 
	    case 0:
		fprintf ( file, "pulse calibration\n" ) ;
		break ;

	    case 1:
		fprintf ( file, "step calibration\n" ) ;
		break ;
	  
	    case 2:
		fprintf ( file, "sinewave calibration\n" ) ;
		break ;

	    case 3:
		fprintf ( file, "random binary calibration\n" ) ;
		break ;
	}
    }
    voltage = (stat>>10) & 0x3f ;
    fprintf ( file, "Digitizer Voltage: 0x%x\n", voltage ) ;
}

static void
showPktChannelSpecific(FILE *file, PktChannel *achan, Packet *pkt) 
{
    if ( strncmp(pkt->parts.src_suffix, "CD1", 3) == 0 ) { 
	switch ( achan->iuser1 ) { 
	    case 1:
		fprintf ( file, "compressed, " ) ;
		break ;

	    case 0:
		fprintf ( file, "not compressed, " ) ;
		break ;

	    default:
		fprintf ( file, "invalid compression flag: %d, ", achan->iuser1);
		break ;
	}

	switch ( achan->iuser2 ) { 
	    case 1:
		fprintf ( file, "signed:\n" ) ;
		hexdump ( file, achan->cuser2, 40 ) ;
		break ;

	    case 0:
		fprintf ( file, "not signed\n" ) ;
		break ;

	    default:
		fprintf ( file, "invalid signed flag: %d\n", achan->iuser2);
		break ;
	}

	show_CD1S_status ( file, achan->cuser1 ) ; 
    }
}


static void
chanshow ( FILE *file, int chan, PktChannel *achan, int samples, Packet *pkt ) 
{
    int i ;
    char *s, *s2 ;
    int *data ;
    char *fmt1 = "%a %Y-%j %b %d %k:%M:%S.%s", 
    	 *fmt2 = "%k:%M:%S.%s" ;

    fprintf ( file, "    %2d : %-8s %-8s %-8s %-8s %6.3f/s calib=%12.8g calper=%6.3f segtype=%s %d samps %s - %s\n",
	chan, 
	achan->net, 
	achan->sta, 
	achan->chan, 
	achan->loc, 
	achan->samprate, 
	achan->calib,
	achan->calper,
	achan->segtype,
	achan->nsamp, 
	s=epoch2str(achan->time, fmt1), 
	s2=epoch2str(SAMP2TIME(achan->time, achan->samprate, achan->nsamp),fmt2)
	) ;
    free(s) ;
    free(s2) ;
    showPktChannelSpecific(file, achan, pkt) ;
    if (samples != 0) {
	int n ; 
	if ( samples < 0 ) { 
	    n = achan->nsamp ; 
	} else { 
	    n = samples ; 
	}
	data = achan->data ;
	fprintf ( file, "\t" ) ;
	for (i=0 ; i<n ; i++ ) {
	    if ( i>0 && (i % 10) == 0 ) fprintf ( file, "\n\t" ) ; 
	    fprintf ( file, " %8d", data[i] ) ; 
	}
	fprintf ( file, "\n" ) ; 
    }
}


void
showPkt ( int pktid, char *srcname, double pkttime, char *packet, int nbytes, 
	FILE *file, int mode )  
{
    Packet *unstuffed=0 ;
    int i, samples = -1 ;
    PktChannel *achan ;
    char *s ;
    int           (*method)() ;

    elog_debug ( 0, "#%d '%s': %s : %d bytes\n", 
	    pktid, srcname, s=strydtime(pkttime), nbytes ) ; 
    free(s) ;

    switch (mode) {
	case PKT_DUMP:
	    fprintf ( file, "%s %s %s %d %.7f\n", 
		    "#orb", "x", srcname, nbytes, pkttime ) ;
	    hexdump ( file, packet, nbytes ) ; 
	    break ;

	default:
	case PKT_NOSAMPLES:
	case PKT_UNSTUFF:
	case PKT_PEEK:
	    switch ( unstuffPkt ( srcname, pkttime, packet, nbytes, &unstuffed ) ) { 
		case Pkt_wf :
		    switch (mode) { 
			case PKT_PEEK:
			    samples = 10 ; 
			    break; 
			case PKT_NOSAMPLES:
			    samples = 0 ; 
			    break; 
			case PKT_UNSTUFF:
			    samples = -1 ; 
			    break ;
		    }
		    showPktSpecific(file, unstuffed) ;
		    for ( i=0 ; i<unstuffed->nchannels ; i++ ) {
			achan = (PktChannel *) gettbl(unstuffed->channels, i) ;
			chanshow ( file, i, achan, samples, unstuffed ) ; 
		    }
		    break ;

		case Pkt_st:			       /* status packet */
		    fprintf ( file, "    status packet\n" ) ;
		    hexdump ( file, packet, nbytes ) ; 
		    break;

		case Pkt_stash:			       /* stash packet */
		    fprintf ( file, "    stash packet\n" ) ;
		    show_stash ( file, unstuffed ) ; 
		    break;
	 
		 case Pkt_db:
		    fprintf(file, "    database row\n" ) ;
		    show_db ( file, unstuffed->db ) ; 
		    if ( unstuffed->dfile ) { 
			fprintf (file, "dfile:\n" ) ;
			fputs (unstuffed->dfile, file ) ; 
			fprintf (file, "\n" ) ;
		    }
		    break;

		case Pkt_pf:
		    fprintf ( file, "    parameter file data packet\n" ) ;
		    pfout(file, unstuffed->pf) ;
		    break;

		case Pkt_cn:
		    fprintf ( file, "    control message packet\n" ) ;
		    pfout(file, unstuffed->pf) ;
		    break;

		case Pkt_rw:
		    fprintf ( file, "    raw data packet\n" ) ;
		    if ( mode == PKT_PEEK ) { 
			hexdump ( file, packet, 48 ) ; 
		    } else { 
			hexdump ( file, packet, nbytes ) ; 
		    }
		    break;

		case Pkt_ch:
		    fprintf ( file, "    string packet\n" ) ; 
		    fprintf ( file, "    '%s'\n", unstuffed->string ) ;
		    break;

		case Pkt_fl:
		    show_fl (pktid, srcname, pkttime, packet, nbytes, unstuffed, file, mode ) ;
		    break;

		default: 
		    method = (int (*)()) usermethod("showPkt_User") ;
		    if ( method != 0 ) {
			method (pktid, srcname, pkttime, packet, nbytes, unstuffed, file, mode) ;
		    } else {
			int n ;
			n = elog_mark() ; 
			if ( n > 0 ) { 
			    complain ( 0, "*** couldn't unstuff packet ***\n" ) ; 
			} else { 
			    fprintf ( file, "*** couldn't unstuff packet ***\n" ) ; 
			}
			fprintf ( file, "%s %s %s %d %.7f\n", 
				"#orb", "x", srcname, nbytes, pkttime ) ;
			if ( mode == PKT_PEEK ) { 
			    hexdump (file, packet, 48) ; 
			}
		    }
		    break ;
	    }
	    freePkt(unstuffed) ;
	    break ;

	case PKT_TERSE:
	    break ;
    }
}

